"""Performance benchmarks for key CLI operations.

Uses pytest-benchmark to track execution time and catch performance regressions.
Run with: pytest tests/test_performance.py --benchmark-enable
"""

from __future__ import annotations

from pathlib import Path

import pytest

from cihub.cli import main


@pytest.fixture
def python_repo(tmp_path: Path) -> Path:
    """Create a minimal Python repo fixture."""
    (tmp_path / "pyproject.toml").write_text("[project]\nname = 'test-project'\nversion = '1.0.0'\n")
    (tmp_path / "src").mkdir()
    (tmp_path / "src" / "__init__.py").write_text("")
    (tmp_path / "tests").mkdir()
    (tmp_path / "tests" / "__init__.py").write_text("")
    return tmp_path


@pytest.fixture
def java_repo(tmp_path: Path) -> Path:
    """Create a minimal Java repo fixture."""
    pom = """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.example</groupId>
    <artifactId>test-project</artifactId>
    <version>1.0.0</version>
</project>
"""
    (tmp_path / "pom.xml").write_text(pom)
    (tmp_path / "src" / "main" / "java").mkdir(parents=True)
    (tmp_path / "src" / "test" / "java").mkdir(parents=True)
    return tmp_path


class TestDetectPerformance:
    """Benchmark detect command - should be fast."""

    def test_detect_python_speed(self, benchmark, python_repo: Path) -> None:
        """Detect Python repo should be fast (<100ms)."""

        def run():
            return main(["detect", "--repo", str(python_repo)])

        result = benchmark(run)
        assert result == 0

    def test_detect_java_speed(self, benchmark, java_repo: Path) -> None:
        """Detect Java repo should be fast (<100ms)."""

        def run():
            return main(["detect", "--repo", str(java_repo)])

        result = benchmark(run)
        assert result == 0


class TestInitPerformance:
    """Benchmark init command - should be reasonably fast."""

    def test_init_python_speed(self, benchmark, tmp_path: Path) -> None:
        """Init Python repo should complete quickly."""
        counter = [0]

        def run():
            # Use unique path for each run
            repo = tmp_path / f"repo_{counter[0]}"
            repo.mkdir()
            (repo / "pyproject.toml").write_text("[project]\n")
            counter[0] += 1
            return main(
                [
                    "init",
                    "--repo",
                    str(repo),
                    "--language",
                    "python",
                    "--owner",
                    "test",
                    "--name",
                    "repo",
                    "--branch",
                    "main",
                    "--apply",
                ]
            )

        result = benchmark(run)
        assert result == 0

    def test_init_java_speed(self, benchmark, tmp_path: Path) -> None:
        """Init Java repo should complete quickly."""
        counter = [0]

        def run():
            repo = tmp_path / f"repo_{counter[0]}"
            repo.mkdir()
            (repo / "pom.xml").write_text("<project></project>")
            counter[0] += 1
            return main(
                [
                    "init",
                    "--repo",
                    str(repo),
                    "--language",
                    "java",
                    "--owner",
                    "test",
                    "--name",
                    "repo",
                    "--branch",
                    "main",
                    "--apply",
                ]
            )

        result = benchmark(run)
        assert result == 0


class TestValidatePerformance:
    """Benchmark validate command."""

    def test_validate_speed(self, benchmark, python_repo: Path) -> None:
        """Validate should be fast for simple configs."""
        # Create config
        (python_repo / ".ci-hub.yml").write_text("language: python\nrepo:\n  owner: test\n  name: repo\n")

        def run():
            return main(["validate", "--repo", str(python_repo)])

        result = benchmark(run)
        # May succeed or fail, just checking performance
        assert result in (0, 1)


class TestScaffoldPerformance:
    """Benchmark scaffold command - file generation."""

    def test_scaffold_python_speed(self, benchmark, tmp_path: Path) -> None:
        """Scaffold Python project should be fast."""
        counter = [0]

        def run():
            path = tmp_path / f"scaffold_{counter[0]}"
            counter[0] += 1
            return main(["scaffold", "python-pyproject", str(path)])

        result = benchmark(run)
        assert result == 0


class TestDiscoverPerformance:
    """Benchmark discover command - matrix generation."""

    def test_discover_speed(self, benchmark) -> None:
        """Discover should complete in reasonable time."""

        def run():
            return main(["discover"])

        result = benchmark(run)
        assert result in (0, 1)


class TestConfigLoadPerformance:
    """Benchmark config loading operations."""

    def test_config_merge_speed(self, benchmark) -> None:
        """Config merging should be fast."""
        from cihub.config.merge import deep_merge

        base = {
            "language": "python",
            "python": {"version": "3.12", "tools": {"pytest": {"enabled": True}}},
            "thresholds": {"coverage_min": 70},
        }
        override = {
            "python": {"tools": {"ruff": {"enabled": True}}},
            "thresholds": {"coverage_min": 80},
        }

        def run():
            return deep_merge(base.copy(), override)

        result = benchmark(run)
        assert "python" in result

    def test_config_normalize_speed(self, benchmark) -> None:
        """Config normalization should be fast."""
        from cihub.config.normalize import normalize_config

        config = {
            "language": "python",
            "python": {"tools": {"pytest": True, "ruff": False}},
        }

        def run():
            return normalize_config(config.copy())

        result = benchmark(run)
        assert "python" in result


class TestReportBuildPerformance:
    """Benchmark report building operations."""

    def test_report_validation_speed(self, benchmark, tmp_path: Path) -> None:
        """Report validation should be fast."""
        import json

        report = tmp_path / "report.json"
        report.write_text(
            json.dumps(
                {
                    "schema_version": "2.0",
                    "repository": "test/repo",
                    "branch": "main",
                    "python_version": "3.12",
                    "results": {
                        "coverage": 80,
                        "tests_passed": 100,
                        "tests_failed": 0,
                    },
                    "tool_metrics": {},
                    "tools_configured": {"pytest": True},
                    "tools_ran": {"pytest": True},
                    "tools_success": {"pytest": True},
                }
            )
        )

        def run():
            return main(["report", "validate", "--report", str(report)])

        result = benchmark(run)
        assert result == 0


# Performance thresholds (optional - uncomment to enforce)
# @pytest.mark.benchmark(min_time=0.1, max_time=0.5, min_rounds=5)
# def test_detect_under_threshold(benchmark, python_repo: Path) -> None:
#     """Detect must complete under 500ms."""
#     result = benchmark(lambda: main(["detect", "--repo", str(python_repo)]))
#     assert result == 0



"""Tests for CLI environment overrides in cihub.services.ci_engine."""

from cihub.services.ci_engine import _apply_env_overrides, _parse_env_bool


def _base_config(language: str) -> dict:
    return {
        "language": language,
        language: {"tools": {}},
    }


def test_parse_env_bool_values() -> None:
    assert _parse_env_bool("true") is True
    assert _parse_env_bool("1") is True
    assert _parse_env_bool("yes") is True
    assert _parse_env_bool("on") is True
    assert _parse_env_bool("false") is False
    assert _parse_env_bool("0") is False
    assert _parse_env_bool("no") is False
    assert _parse_env_bool("off") is False
    assert _parse_env_bool("maybe") is None
    assert _parse_env_bool(None) is None


def test_apply_env_overrides_python_tools() -> None:
    config = _base_config("python")
    config["python"]["tools"]["pytest"] = {"enabled": True}
    config["python"]["tools"]["ruff"] = False
    problems: list[dict[str, object]] = []
    env = {
        "CIHUB_RUN_PYTEST": "false",
        "CIHUB_RUN_RUFF": "true",
    }

    _apply_env_overrides(config, "python", env, problems)

    assert config["python"]["tools"]["pytest"]["enabled"] is False
    assert config["python"]["tools"]["ruff"]["enabled"] is True
    assert problems == []


def test_apply_env_overrides_java_tools() -> None:
    config = _base_config("java")
    config["java"]["tools"]["jacoco"] = True
    problems: list[dict[str, object]] = []
    env = {"CIHUB_RUN_JACOCO": "false"}

    _apply_env_overrides(config, "java", env, problems)

    assert config["java"]["tools"]["jacoco"]["enabled"] is False
    assert problems == []


def test_apply_env_overrides_invalid_value_warns() -> None:
    config = _base_config("python")
    problems: list[dict[str, object]] = []
    env = {"CIHUB_RUN_PYTEST": "maybe"}

    _apply_env_overrides(config, "python", env, problems)

    assert len(problems) == 1
    assert problems[0]["code"] == "CIHUB-CI-ENV-BOOL"


def test_apply_env_overrides_summary_toggle() -> None:
    config = _base_config("python")
    problems: list[dict[str, object]] = []
    env = {"CIHUB_WRITE_GITHUB_SUMMARY": "false"}

    _apply_env_overrides(config, "python", env, problems)

    assert config["reports"]["github_summary"]["enabled"] is False
    assert problems == []



"""Shared argument factories for CLI parsers.

This module eliminates 36+ duplicate argument definitions by providing
reusable factory functions for common CLI patterns.

Usage:
    from cihub.cli_parsers.common import add_output_args, add_summary_args

    parser = subparsers.add_parser("my-command", help="My command")
    add_output_args(parser)  # Adds --output and --github-output
    add_summary_args(parser)  # Adds --summary

★ Insight ─────────────────────────────────────
Why factory functions instead of inheritance/mixins?
1. argparse doesn't support parser inheritance cleanly
2. Factory functions compose - call multiple for different combos
3. Easy to test: just verify arguments exist on mock parser
─────────────────────────────────────────────────
"""

from __future__ import annotations

from argparse import ArgumentParser


def add_output_args(
    parser: ArgumentParser,
    *,
    output_help: str = "Write outputs to file",
    include_github_output: bool = True,
) -> None:
    """Add standard --output and --github-output arguments.

    Args:
        parser: The argparse parser/subparser to add arguments to.
        output_help: Custom help text for --output (default: "Write outputs to file").
        include_github_output: Whether to include --github-output flag (default: True).

    Example:
        >>> parser = ArgumentParser()
        >>> add_output_args(parser)
        >>> args = parser.parse_args(["--output", "out.txt"])
        >>> args.output
        'out.txt'
    """
    parser.add_argument("--output", help=output_help)
    if include_github_output:
        parser.add_argument(
            "--github-output",
            action="store_true",
            help="Write outputs to GITHUB_OUTPUT",
        )


def add_summary_args(
    parser: ArgumentParser,
    *,
    summary_help: str = "Write summary to file",
) -> None:
    """Add standard --summary argument.

    Args:
        parser: The argparse parser/subparser to add arguments to.
        summary_help: Custom help text for --summary.

    Example:
        >>> parser = ArgumentParser()
        >>> add_summary_args(parser)
        >>> args = parser.parse_args(["--summary", "summary.md"])
        >>> args.summary
        'summary.md'
    """
    parser.add_argument("--summary", help=summary_help)


def add_repo_args(
    parser: ArgumentParser,
    *,
    required: bool = False,
    default: str | None = ".",
    help_text: str = "Path to repo",
) -> None:
    """Add standard --repo argument.

    Args:
        parser: The argparse parser/subparser to add arguments to.
        required: Whether --repo is required (default: False).
        default: Default value when not required (default: ".").
        help_text: Custom help text.

    Example:
        >>> parser = ArgumentParser()
        >>> add_repo_args(parser)
        >>> args = parser.parse_args([])
        >>> args.repo
        '.'
    """
    if required:
        parser.add_argument("--repo", required=True, help=help_text)
    else:
        full_help = f"{help_text} (default: {default})" if default else help_text
        parser.add_argument("--repo", default=default, help=full_help)


def add_report_args(
    parser: ArgumentParser,
    *,
    required: bool = False,
    default: str | None = None,
    help_text: str = "Path to report.json",
) -> None:
    """Add standard --report argument.

    Args:
        parser: The argparse parser/subparser to add arguments to.
        required: Whether --report is required (default: False).
        default: Default value when not required.
        help_text: Custom help text.

    Example:
        >>> parser = ArgumentParser()
        >>> add_report_args(parser, required=True)
        >>> # --report is now required
    """
    if required:
        parser.add_argument("--report", required=True, help=help_text)
    else:
        parser.add_argument("--report", default=default, help=help_text)


def add_path_args(
    parser: ArgumentParser,
    *,
    default: str = "repo",
    help_text: str = "Repo path",
) -> None:
    """Add standard --path argument (for repo checkout paths).

    Args:
        parser: The argparse parser/subparser to add arguments to.
        default: Default value (default: "repo").
        help_text: Custom help text.
    """
    parser.add_argument("--path", default=default, help=help_text)


def add_output_dir_args(
    parser: ArgumentParser,
    *,
    default: str = ".",
    help_text: str = "Output directory",
) -> None:
    """Add standard --output-dir argument.

    Args:
        parser: The argparse parser/subparser to add arguments to.
        default: Default value (default: ".").
        help_text: Custom help text.
    """
    parser.add_argument("--output-dir", default=default, help=help_text)


# =============================================================================
# Composite Helpers (common combinations)
# =============================================================================


def add_ci_output_args(parser: ArgumentParser) -> None:
    """Add both --output and --summary for CI commands.

    This is a common pattern for hub-ci subcommands that produce
    both machine-readable outputs and human-readable summaries.
    """
    add_output_args(parser)
    add_summary_args(parser)


def add_tool_runner_args(
    parser: ArgumentParser,
    *,
    path_default: str = "repo",
) -> None:
    """Add standard arguments for tool runner commands.

    Adds: --path, --output, --github-output

    Args:
        parser: The argparse parser/subparser to add arguments to.
        path_default: Default value for --path.
    """
    add_path_args(parser, default=path_default)
    add_output_args(parser)


# =============================================================================
# See Also References
# =============================================================================

# Mapping of command names to related commands for help epilogs
SEE_ALSO: dict[str, list[str]] = {
    # Core workflow
    "detect": ["init", "preflight", "scaffold"],
    "preflight": ["detect", "check", "smoke"],
    "doctor": ["preflight", "check"],
    "scaffold": ["smoke", "init", "detect"],
    "smoke": ["scaffold", "check", "ci"],
    "check": ["smoke", "verify", "ci", "preflight"],
    "verify": ["check", "sync-templates", "ci"],
    "ci": ["check", "run", "report", "triage"],
    "run": ["ci", "tool list"],
    # Reporting & triage
    "report": ["ci", "triage"],
    "triage": ["ci", "report"],
    # Registry management
    "registry": ["profile", "tool", "threshold", "repo"],
    "profile": ["registry", "tool", "threshold"],
    "tool": ["registry", "profile", "run"],
    "threshold": ["registry", "profile"],
    "repo": ["registry", "config"],
    # Init & setup
    "init": ["detect", "scaffold", "update", "setup"],
    "update": ["init", "validate"],
    "validate": ["init", "check"],
    "setup": ["init", "setup-secrets", "setup-nvd"],
    "setup-secrets": ["setup", "setup-nvd"],
    "setup-nvd": ["setup", "setup-secrets"],
    # Documentation
    "docs": ["adr"],
    "adr": ["docs"],
    # Hub management
    "hub": ["registry", "hub-ci"],
    "hub-ci": ["hub", "ci", "triage"],
    "config": ["registry", "hub"],
    # Fix commands
    "fix": ["fix-pom", "fix-deps", "fix-gradle"],
    "fix-pom": ["fix", "fix-deps"],
    "fix-deps": ["fix", "fix-pom"],
    "fix-gradle": ["fix"],
}


def see_also_epilog(command: str) -> str | None:
    """Generate a See Also epilog for a command.

    Args:
        command: The command name to get related commands for.

    Returns:
        Formatted epilog string, or None if no related commands.

    Example:
        >>> see_also_epilog("smoke")
        'See also: scaffold, check, ci'
    """
    related = SEE_ALSO.get(command)
    if not related:
        return None
    return f"See also: {', '.join(related)}"



"""Java tool selection prompts."""

from __future__ import annotations

from copy import deepcopy
from typing import Any

import questionary

from cihub.tools.registry import JAVA_TOOLS, get_custom_tools_from_config
from cihub.wizard.core import _check_cancelled
from cihub.wizard.styles import get_style


def configure_java_tools(defaults: dict) -> dict[str, Any]:
    """Prompt to enable/disable Java tools.

    Sources tool list from cihub.tools.registry (CLI source of truth).
    Also includes any custom x-* tools defined in the config.

    Args:
        defaults: Defaults config (expects java.tools).

    Returns:
        Tool config dict with updated enabled flags.
    """
    raw_tools = deepcopy(defaults.get("java", {}).get("tools", {}))
    tools: dict[str, Any] = raw_tools if isinstance(raw_tools, dict) else {}

    # Get all tools: built-in from registry + custom from config
    all_tools = list(JAVA_TOOLS)
    custom_tools = get_custom_tools_from_config(defaults, "java")
    all_tools.extend(sorted(custom_tools.keys()))

    for tool in all_tools:
        tool_cfg = tools.get(tool)
        if not isinstance(tool_cfg, dict):
            # Initialize config for tools not yet in defaults
            tool_cfg = {"enabled": False}
            tools[tool] = tool_cfg
        enabled = tool_cfg.get("enabled", False)
        # Show custom tool indicator
        label = f"Enable {tool}?" if tool in JAVA_TOOLS else f"Enable {tool} (custom)?"
        answer: bool = _check_cancelled(
            questionary.confirm(
                label,
                default=bool(enabled),
                style=get_style(),
            ).ask(),
            f"{tool} toggle",
        )
        tool_cfg["enabled"] = bool(answer)
    return tools



"""Registry command package for centralized repo configuration."""

from __future__ import annotations

import argparse

from cihub.exit_codes import EXIT_USAGE
from cihub.types import CommandResult

# Import utility function first (no circular dependency)
from ._utils import _derive_hub_root_from_configs_dir

# Import handlers from submodules
from .io import _cmd_export, _cmd_import
from .modify import _cmd_add, _cmd_remove, _cmd_set
from .query import _cmd_list, _cmd_show
from .sync import _cmd_bootstrap, _cmd_diff, _cmd_sync


def cmd_registry(args: argparse.Namespace) -> CommandResult:
    """Dispatch registry subcommands."""
    subcommand: str | None = getattr(args, "subcommand", None)

    handlers = {
        "list": _cmd_list,
        "show": _cmd_show,
        "set": _cmd_set,
        "diff": _cmd_diff,
        "sync": _cmd_sync,
        "add": _cmd_add,
        "remove": _cmd_remove,
        "bootstrap": _cmd_bootstrap,
        "export": _cmd_export,
        "import": _cmd_import,
    }

    handler = handlers.get(subcommand or "")
    if handler is None:
        return CommandResult(
            exit_code=EXIT_USAGE,
            summary=f"Unknown registry subcommand: {subcommand}",
            problems=[
                {
                    "severity": "error",
                    "message": f"Unknown subcommand '{subcommand}'",
                    "code": "CIHUB-REGISTRY-UNKNOWN-SUBCOMMAND",
                }
            ],
        )

    return handler(args)


# Export all handlers for backward compatibility and direct access
__all__ = [
    "cmd_registry",
    "_derive_hub_root_from_configs_dir",
    "_cmd_list",
    "_cmd_show",
    "_cmd_set",
    "_cmd_add",
    "_cmd_remove",
    "_cmd_diff",
    "_cmd_sync",
    "_cmd_bootstrap",
    "_cmd_export",
    "_cmd_import",
]



import argparse
from pathlib import Path

import pytest

from cihub.commands.scaffold import cmd_scaffold
from cihub.types import CommandResult


class TestScaffoldPython:
    """Tests for Python scaffold types."""

    def test_scaffold_python_pyproject(self, tmp_path: Path) -> None:
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="python-pyproject",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / "pyproject.toml").exists()
        assert (dest / "tests" / "test_app.py").exists()

    def test_scaffold_python_setup(self, tmp_path: Path) -> None:
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="python-setup",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / "setup.py").exists()
        assert (dest / "tests" / "test_core.py").exists()

    def test_scaffold_python_src_layout(self, tmp_path: Path) -> None:
        """Test python-src-layout scaffold with src/ directory structure."""
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="python-src-layout",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / "pyproject.toml").exists()
        assert (dest / "src").exists()
        assert (dest / "src" / "cihub_src_sample" / "__init__.py").exists()
        assert (dest / "src" / "cihub_src_sample" / "core.py").exists()
        assert (dest / "tests" / "test_core.py").exists()


class TestScaffoldJava:
    """Tests for Java scaffold types."""

    def test_scaffold_java_maven(self, tmp_path: Path) -> None:
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="java-maven",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / "pom.xml").exists()
        assert (dest / "src" / "main" / "java" / "com" / "cihub" / "App.java").exists()
        assert (dest / "src" / "test" / "java" / "com" / "cihub" / "AppTest.java").exists()

    def test_scaffold_java_gradle(self, tmp_path: Path) -> None:
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="java-gradle",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / "build.gradle").exists()
        assert (dest / "settings.gradle").exists()
        assert (dest / "src" / "main" / "java" / "com" / "cihub" / "App.java").exists()
        assert (dest / "src" / "test" / "java" / "com" / "cihub" / "AppTest.java").exists()

    def test_scaffold_java_multi_module(self, tmp_path: Path) -> None:
        """Test java-multi-module scaffold with parent POM and child modules."""
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="java-multi-module",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        # Parent POM
        assert (dest / "pom.xml").exists()
        # Child modules
        assert (dest / "module-core" / "pom.xml").exists()
        assert (dest / "module-app" / "pom.xml").exists()
        # Source files
        assert (dest / "module-core" / "src" / "main" / "java" / "com" / "cihub" / "core" / "Calculator.java").exists()
        assert (dest / "module-app" / "src" / "main" / "java" / "com" / "cihub" / "app" / "App.java").exists()


class TestScaffoldMonorepo:
    """Tests for monorepo scaffold type."""

    def test_scaffold_monorepo(self, tmp_path: Path) -> None:
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type="monorepo",
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        # Monorepo should have both java/ and python/ subdirs
        assert (dest / "java").exists() or (dest / "python").exists(), "Monorepo should have language subdirs"


class TestScaffoldList:
    """Tests for scaffold --list."""

    def test_scaffold_list_json(self) -> None:
        args = argparse.Namespace(list=True, type=None, path=None, force=False, json=True)
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert "fixtures" in result.data

    def test_scaffold_list_contains_all_types(self) -> None:
        """Ensure all scaffold types are discoverable."""
        args = argparse.Namespace(list=True, type=None, path=None, force=False, json=True)
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        fixtures = result.data.get("fixtures", [])
        types = [f["type"] for f in fixtures]
        # All expected types should be present (all 7 scaffold types)
        expected = [
            "python-pyproject",
            "python-setup",
            "python-src-layout",
            "java-maven",
            "java-gradle",
            "java-multi-module",
            "monorepo",
        ]
        for exp in expected:
            assert exp in types, f"Missing scaffold type: {exp}"


class TestScaffoldValidation:
    """Validate that scaffold templates produce valid project structures."""

    @pytest.mark.parametrize(
        "scaffold_type,build_file",
        [
            ("python-pyproject", "pyproject.toml"),
            ("python-setup", "setup.py"),
            ("python-src-layout", "pyproject.toml"),
            ("java-maven", "pom.xml"),
            ("java-gradle", "build.gradle"),
            ("java-multi-module", "pom.xml"),
        ],
    )
    def test_scaffold_produces_buildable_project(self, tmp_path: Path, scaffold_type: str, build_file: str) -> None:
        """Each scaffold type should produce a project with valid build config."""
        dest = tmp_path / "fixture"
        args = argparse.Namespace(
            list=False,
            type=scaffold_type,
            path=str(dest),
            force=False,
            json=True,
        )
        result = cmd_scaffold(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert (dest / build_file).exists(), f"Missing {build_file} for {scaffold_type}"

        # Verify build file is not empty and contains expected content
        content = (dest / build_file).read_text()
        assert len(content) > 50, f"Build file {build_file} seems too small"



"""Java Gradle file parsing and manipulation utilities.

This module provides functions for parsing Gradle build files, detecting
missing plugins, and manipulating build.gradle content.
"""

from __future__ import annotations

import re
import textwrap
from pathlib import Path
from typing import Any

from cihub.utils.paths import hub_root

# ============================================================================
# Constants
# ============================================================================

# Map of tool name to Gradle plugin ID
GRADLE_TOOL_PLUGINS = {
    "jacoco": "jacoco",
    "checkstyle": "checkstyle",
    "spotbugs": "com.github.spotbugs",
    "pmd": "pmd",
    "owasp": "org.owasp.dependencycheck",
    "pitest": "info.solidsoft.pitest",
}


# ============================================================================
# Tool Configuration
# ============================================================================


def get_gradle_tool_flags(config: dict[str, Any]) -> dict[str, bool]:
    """Get enabled status for Java tools per ADR-0017 defaults."""
    tools = config.get("java", {}).get("tools", {})
    # Defaults per ADR-0017: Core tools enabled, expensive tools disabled
    defaults = {
        "jacoco": True,
        "checkstyle": True,
        "spotbugs": True,
        "pmd": True,
        "owasp": True,
        "pitest": True,
        "jqwik": False,  # opt-in
        "semgrep": False,  # expensive
        "trivy": False,  # expensive
        "codeql": False,  # expensive
        "docker": False,  # requires Dockerfile
    }
    enabled: dict[str, bool] = {}
    for tool, default in defaults.items():
        enabled[tool] = tools.get(tool, {}).get("enabled", default)
    return enabled


# ============================================================================
# Gradle Parsing
# ============================================================================


def parse_gradle_plugins(build_path: Path) -> tuple[set[str], str | None]:
    """Parse plugins from a Gradle build file.

    Returns:
        Tuple of (plugin_ids, error)
    """
    if not build_path.exists():
        return set(), f"build.gradle not found: {build_path}"

    try:
        content = build_path.read_text(encoding="utf-8")
    except (OSError, UnicodeDecodeError) as exc:
        return set(), f"Error reading {build_path}: {exc}"

    plugins: set[str] = set()

    # Match plugins block
    plugins_block = re.search(r"plugins\s*\{([^}]*)\}", content, re.DOTALL)
    if plugins_block:
        block_content = plugins_block.group(1)
        # Match plugin declarations:
        # id 'plugin-id'
        # id 'plugin-id' version 'x.y.z'
        # id("plugin-id")
        # id("plugin-id") version "x.y.z"
        for match in re.finditer(r"id\s*[('\"]([^'\"]+)['\"]", block_content):
            plugins.add(match.group(1))

    # Also check for apply plugin syntax (legacy)
    for match in re.finditer(r"apply\s+plugin:\s*['\"]([^'\"]+)['\"]", content):
        plugins.add(match.group(1))

    return plugins, None


def find_plugins_block(content: str) -> tuple[int, int] | None:
    """Find the plugins block in Gradle content.

    Returns:
        Tuple of (start_index, end_index) or None if not found
    """
    match = re.search(r"plugins\s*\{", content)
    if not match:
        return None

    # Find matching closing brace
    start = match.end()
    depth = 1
    for i, char in enumerate(content[start:], start):
        if char == "{":
            depth += 1
        elif char == "}":
            depth -= 1
            if depth == 0:
                return (match.start(), i + 1)

    return None


def find_insertion_point_for_config(content: str) -> int:
    """Find the best place to insert configuration blocks.

    Returns the index after the plugins block, or at the end of file.
    """
    plugins_span = find_plugins_block(content)
    if plugins_span:
        # Insert after plugins block
        end_idx = plugins_span[1]
        # Skip any trailing whitespace/newlines
        while end_idx < len(content) and content[end_idx] in " \t\n":
            end_idx += 1
        return end_idx

    # No plugins block, find after repositories or at start
    repos_match = re.search(r"repositories\s*\{[^}]*\}", content, re.DOTALL)
    if repos_match:
        end_idx = repos_match.end()
        while end_idx < len(content) and content[end_idx] in " \t\n":
            end_idx += 1
        return end_idx

    # Insert at beginning (after any comments)
    first_code = re.search(r"^\s*[^/\s*]", content, re.MULTILINE)
    return first_code.start() if first_code else 0


# ============================================================================
# Warning Collection
# ============================================================================


def collect_gradle_warnings(repo_path: Path, config: dict[str, Any]) -> tuple[list[str], list[str]]:
    """Collect warnings about missing plugins in Gradle build file.

    Returns:
        Tuple of (warnings, missing_plugin_ids)
    """
    warnings: list[str] = []
    missing_plugins: list[str] = []

    subdir = config.get("repo", {}).get("subdir") or ""
    root_path = repo_path / subdir if subdir else repo_path
    build_path = root_path / "build.gradle"

    if not build_path.exists():
        # Check for Kotlin DSL
        kotlin_path = root_path / "build.gradle.kts"
        if kotlin_path.exists():
            warnings.append("Kotlin DSL (build.gradle.kts) detected; fix-gradle only supports Groovy DSL")
            return warnings, missing_plugins
        warnings.append("build.gradle not found")
        return warnings, missing_plugins

    build_tool = config.get("java", {}).get("build_tool", "maven")
    if build_tool != "gradle":
        return warnings, missing_plugins

    declared_plugins, error = parse_gradle_plugins(build_path)
    if error:
        warnings.append(error)
        return warnings, missing_plugins

    tool_flags = get_gradle_tool_flags(config)

    for tool, enabled in tool_flags.items():
        if tool not in GRADLE_TOOL_PLUGINS or not enabled:
            continue
        plugin_id = GRADLE_TOOL_PLUGINS[tool]
        if plugin_id not in declared_plugins:
            warnings.append(f"build.gradle: missing plugin for enabled tool '{tool}' ({plugin_id})")
            missing_plugins.append(plugin_id)

    return warnings, missing_plugins


# ============================================================================
# Snippet Loading
# ============================================================================


def load_gradle_plugin_snippets() -> dict[str, str]:
    """Load plugin declaration snippets from the templates directory.

    Returns:
        Dict mapping plugin_id to plugin declaration line
    """
    snippets_path = hub_root() / "templates" / "java" / "gradle-plugins.groovy"
    if not snippets_path.exists():
        return {}

    content = snippets_path.read_text(encoding="utf-8")
    snippets: dict[str, str] = {}

    # Parse // @plugin:<plugin-id> markers
    current_plugin: str | None = None
    lines = content.splitlines()

    for line in lines:
        marker_match = re.match(r"^\s*//\s*@plugin:(.+)$", line)
        if marker_match:
            current_plugin = marker_match.group(1).strip()
            continue

        if current_plugin and line.strip().startswith("id "):
            # Found the plugin declaration line
            snippets[current_plugin] = line.strip()
            current_plugin = None

    return snippets


def load_gradle_config_snippets() -> dict[str, str]:
    """Load plugin configuration snippets from the templates directory.

    Returns:
        Dict mapping plugin_id to configuration block
    """
    snippets_path = hub_root() / "templates" / "java" / "gradle-plugins.groovy"
    if not snippets_path.exists():
        return {}

    content = snippets_path.read_text(encoding="utf-8")
    snippets: dict[str, str] = {}

    # Parse // @config:<plugin-id> markers
    current_config: str | None = None
    config_lines: list[str] = []
    lines = content.splitlines()

    for line in lines:
        marker_match = re.match(r"^\s*//\s*@config:(.+)$", line)
        if marker_match:
            # Save previous config if any
            if current_config and config_lines:
                snippets[current_config] = "\n".join(config_lines).strip()
            current_config = marker_match.group(1).strip()
            config_lines = []
            continue

        if current_config:
            # Check if we hit a new marker or end
            if line.strip().startswith("// @"):
                if config_lines:
                    snippets[current_config] = "\n".join(config_lines).strip()
                current_config = None
                config_lines = []
            elif line.strip().startswith("/*"):
                # End at comment block
                if config_lines:
                    snippets[current_config] = "\n".join(config_lines).strip()
                current_config = None
                config_lines = []
            elif line.strip() or config_lines:  # Skip leading empty lines
                config_lines.append(line)

    # Save last config
    if current_config and config_lines:
        snippets[current_config] = "\n".join(config_lines).strip()

    return snippets


# ============================================================================
# Build File Modification
# ============================================================================


def insert_plugins_into_gradle(content: str, plugin_ids: list[str], snippets: dict[str, str]) -> tuple[str, bool]:
    """Insert plugin declarations into a Gradle build file.

    Returns:
        Tuple of (updated_content, success)
    """
    if not plugin_ids:
        return content, True

    # Build plugin lines to insert
    plugin_lines = []
    for plugin_id in plugin_ids:
        if plugin_id in snippets:
            plugin_lines.append(f"    {snippets[plugin_id]}")
        else:
            # Fallback: generate basic plugin declaration
            if "." in plugin_id:
                # External plugin needs version (use placeholder)
                plugin_lines.append(f"    id '{plugin_id}' version 'FIXME'")
            else:
                # Core Gradle plugin
                plugin_lines.append(f"    id '{plugin_id}'")

    plugins_block = find_plugins_block(content)

    if plugins_block:
        # Insert into existing plugins block
        start, end = plugins_block
        # Find position just before closing brace
        insert_pos = end - 1
        # Skip whitespace before closing brace
        while insert_pos > start and content[insert_pos - 1] in " \t\n":
            insert_pos -= 1

        # Get indentation of existing plugins
        block_content = content[start:end]
        indent_match = re.search(r"\n(\s+)id\s", block_content)
        if indent_match:
            indent = indent_match.group(1)
            plugin_lines = [line.replace("    ", indent) for line in plugin_lines]

        insert_text = "\n" + "\n".join(plugin_lines)
        updated = content[:insert_pos] + insert_text + content[insert_pos:]
        return updated, True
    else:
        # Create new plugins block at the beginning
        plugins_block_text = "plugins {\n" + "\n".join(plugin_lines) + "\n}\n\n"

        # Find where to insert (after any buildscript block, before other content)
        buildscript_match = re.search(r"buildscript\s*\{", content)
        if buildscript_match:
            # Find end of buildscript block
            depth = 1
            pos = buildscript_match.end()
            while pos < len(content) and depth > 0:
                if content[pos] == "{":
                    depth += 1
                elif content[pos] == "}":
                    depth -= 1
                pos += 1
            # Skip whitespace after buildscript
            while pos < len(content) and content[pos] in " \t\n":
                pos += 1
            updated = content[:pos] + plugins_block_text + content[pos:]
        else:
            # Insert at beginning (after any leading comments)
            first_code = re.search(r"^(?!//)(?!/\*).", content, re.MULTILINE)
            if first_code:
                pos = first_code.start()
                updated = content[:pos] + plugins_block_text + content[pos:]
            else:
                updated = plugins_block_text + content

        return updated, True


def insert_configs_into_gradle(content: str, plugin_ids: list[str], snippets: dict[str, str]) -> tuple[str, bool]:
    """Insert plugin configuration blocks into a Gradle build file.

    Returns:
        Tuple of (updated_content, success)
    """
    if not plugin_ids:
        return content, True

    # Build config blocks to insert
    config_blocks = []
    for plugin_id in plugin_ids:
        if plugin_id in snippets:
            config_blocks.append(snippets[plugin_id])

    if not config_blocks:
        return content, True

    # Find insertion point
    insert_pos = find_insertion_point_for_config(content)

    # Dedent and format config blocks
    formatted_blocks = []
    for block in config_blocks:
        # Ensure consistent formatting
        block = textwrap.dedent(block).strip()
        formatted_blocks.append(block)

    config_text = "\n\n" + "\n\n".join(formatted_blocks) + "\n"

    updated = content[:insert_pos] + config_text + content[insert_pos:]
    return updated, True



"""Services layer for cihub - pure Python APIs returning dataclasses.

This module provides stable APIs for CLI, GUI, and programmatic access.
"""

from cihub.services.aggregation import (
    AggregationResult,
    aggregate_from_dispatch,
    aggregate_from_reports_dir,
)
from cihub.services.ci import CiRunResult, run_ci
from cihub.services.configuration import (
    ConfigLoadResult,
    load_repo_config,
    resolve_tool_path,
    set_nested_value,
    set_tool_enabled,
)
from cihub.services.detection import detect_language, resolve_language
from cihub.services.discovery import (
    DiscoveryFilters,
    DiscoveryResult,
    discover_repositories,
)
from cihub.services.report_summary import (
    ReportSummaryResult,
    render_summary_from_path,
    render_summary_from_report,
)
from cihub.services.report_validator import (
    ValidationResult,
    ValidationRules,
    validate_against_schema,
    validate_report,
    validate_report_file,
)
from cihub.services.triage_service import (
    MultiTriageResult,
    ToolEvidence,
    ToolStatus,
    TriageBundle,
    aggregate_triage_bundles,
    build_tool_evidence,
    detect_flaky_patterns,
    detect_test_count_regression,
    generate_triage_bundle,
    validate_artifact_evidence,
    write_triage_bundle,
)
from cihub.services.types import RepoEntry, ServiceResult

__all__ = [
    # Types
    "ServiceResult",
    "RepoEntry",
    # CI
    "CiRunResult",
    "run_ci",
    # Config
    "ConfigLoadResult",
    "load_repo_config",
    "resolve_tool_path",
    "set_nested_value",
    "set_tool_enabled",
    # Detection
    "detect_language",
    "resolve_language",
    # Discovery
    "DiscoveryFilters",
    "DiscoveryResult",
    "discover_repositories",
    # Validation
    "ValidationRules",
    "ValidationResult",
    "validate_against_schema",
    "validate_report",
    "validate_report_file",
    # Summary
    "ReportSummaryResult",
    "render_summary_from_path",
    "render_summary_from_report",
    # Aggregation
    "AggregationResult",
    "aggregate_from_dispatch",
    "aggregate_from_reports_dir",
    # Triage
    "MultiTriageResult",
    "ToolEvidence",
    "ToolStatus",
    "TriageBundle",
    "aggregate_triage_bundles",
    "build_tool_evidence",
    "detect_flaky_patterns",
    "detect_test_count_regression",
    "generate_triage_bundle",
    "validate_artifact_evidence",
    "write_triage_bundle",
]



"""Codecov upload and file collection helpers."""

from __future__ import annotations

import shutil
from pathlib import Path
from typing import Any

from cihub.utils.exec_utils import (
    TIMEOUT_NETWORK,
    CommandNotFoundError,
    CommandTimeoutError,
    safe_run,
)


def _collect_codecov_files(
    language: str,
    output_dir: Path,
    tool_outputs: dict[str, dict[str, Any]],
) -> list[Path]:
    files: list[Path] = []
    if language == "python":
        coverage_path = tool_outputs.get("pytest", {}).get("artifacts", {}).get("coverage")
        if coverage_path:
            files.append(Path(coverage_path))
        else:
            files.append(output_dir / "coverage.xml")
    elif language == "java":
        jacoco_path = tool_outputs.get("jacoco", {}).get("artifacts", {}).get("report")
        if jacoco_path:
            files.append(Path(jacoco_path))
    return [path for path in files if path and path.exists()]


def _run_codecov_upload(
    files: list[Path],
    fail_ci_on_error: bool,
    problems: list[dict[str, Any]],
) -> None:
    if not files:
        problems.append(
            {
                "severity": "warning",
                "message": "Codecov enabled but no coverage files were found",
                "code": "CIHUB-CI-CODECOV-NO-FILES",
            }
        )
        return
    codecov_bin = shutil.which("codecov")
    if not codecov_bin:
        problems.append(
            {
                "severity": "error" if fail_ci_on_error else "warning",
                "message": "Codecov enabled but uploader not found in PATH",
                "code": "CIHUB-CI-CODECOV-MISSING",
            }
        )
        return
    cmd = [codecov_bin]
    for path in files:
        cmd.extend(["-f", str(path)])
    try:
        proc = safe_run(cmd, timeout=TIMEOUT_NETWORK)
    except CommandNotFoundError:
        problems.append(
            {
                "severity": "error" if fail_ci_on_error else "warning",
                "message": "Codecov uploader not found",
                "code": "CIHUB-CI-CODECOV-MISSING",
            }
        )
        return
    except CommandTimeoutError:
        problems.append(
            {
                "severity": "error" if fail_ci_on_error else "warning",
                "message": "Codecov upload timed out",
                "code": "CIHUB-CI-CODECOV-TIMEOUT",
            }
        )
        return
    if proc.returncode != 0:
        problems.append(
            {
                "severity": "error" if fail_ci_on_error else "warning",
                "message": f"Codecov upload failed: {proc.stderr.strip() or proc.stdout.strip()}",
                "code": "CIHUB-CI-CODECOV-FAILED",
            }
        )



"""Tests for cihub.commands.config_outputs module."""

from __future__ import annotations

import argparse
import os
from pathlib import Path
from typing import Any
from unittest.mock import patch

from cihub.commands.config_outputs import (
    _bool_str,
    _get_int,
    _get_str,
    _get_value,
    _tool_enabled,
    _tool_entry,
    _tool_int,
    cmd_config_outputs,
)
from cihub.exit_codes import EXIT_FAILURE, EXIT_SUCCESS


class TestGetValue:
    """Tests for _get_value helper."""

    def test_returns_default_for_empty_dict(self) -> None:
        assert _get_value({}, ["a", "b"], "default") == "default"

    def test_returns_value_at_path(self) -> None:
        data = {"a": {"b": {"c": "found"}}}
        assert _get_value(data, ["a", "b", "c"], "default") == "found"

    def test_returns_default_when_path_not_found(self) -> None:
        data = {"a": {"b": "value"}}
        assert _get_value(data, ["a", "c"], "default") == "default"

    def test_returns_default_when_cursor_is_not_dict(self) -> None:
        data = {"a": "not_a_dict"}
        assert _get_value(data, ["a", "b"], "default") == "default"

    def test_returns_default_for_none_value(self) -> None:
        data: dict[str, Any] = {"a": None}
        assert _get_value(data, ["a"], "default") == "default"

    def test_returns_zero_if_actual_value(self) -> None:
        data = {"a": 0}
        assert _get_value(data, ["a"], 999) == 0


class TestGetStr:
    """Tests for _get_str helper."""

    def test_returns_string_value(self) -> None:
        data = {"level1": {"level2": "mystring"}}
        assert _get_str(data, ["level1", "level2"], "default") == "mystring"

    def test_converts_int_to_str(self) -> None:
        data = {"version": 123}
        assert _get_str(data, ["version"], "0") == "123"

    def test_returns_default_when_not_found(self) -> None:
        assert _get_str({}, ["missing"], "fallback") == "fallback"


class TestGetInt:
    """Tests for _get_int helper."""

    def test_returns_int_value(self) -> None:
        data = {"threshold": 85}
        assert _get_int(data, ["threshold"], 70) == 85

    def test_converts_str_to_int(self) -> None:
        data = {"threshold": "90"}
        assert _get_int(data, ["threshold"], 70) == 90

    def test_returns_default_for_non_numeric(self) -> None:
        data = {"threshold": "not_a_number"}
        assert _get_int(data, ["threshold"], 70) == 70

    def test_returns_default_when_missing(self) -> None:
        assert _get_int({}, ["missing"], 50) == 50

    def test_returns_default_for_none(self) -> None:
        data: dict[str, Any] = {"threshold": None}
        assert _get_int(data, ["threshold"], 42) == 42


class TestBoolStr:
    """Tests for _bool_str helper."""

    def test_returns_true_for_true(self) -> None:
        assert _bool_str(True) == "true"

    def test_returns_false_for_false(self) -> None:
        assert _bool_str(False) == "false"


class TestToolEntry:
    """Tests for _tool_entry helper."""

    def test_returns_tool_config(self) -> None:
        config = {"python": {"tools": {"pytest": {"enabled": True, "min_coverage": 80}}}}
        result = _tool_entry(config, "python", "pytest")
        assert result == {"enabled": True, "min_coverage": 80}

    def test_returns_empty_dict_for_missing_language(self) -> None:
        config: dict[str, Any] = {}
        assert _tool_entry(config, "python", "pytest") == {}

    def test_returns_empty_dict_for_missing_tools(self) -> None:
        config = {"python": {}}
        assert _tool_entry(config, "python", "pytest") == {}

    def test_returns_empty_dict_for_non_dict_tools(self) -> None:
        config = {"python": {"tools": "not_a_dict"}}
        assert _tool_entry(config, "python", "pytest") == {}

    def test_returns_empty_dict_for_missing_tool(self) -> None:
        config = {"python": {"tools": {"ruff": True}}}
        result = _tool_entry(config, "python", "pytest")
        assert result == {}


class TestToolEnabled:
    """Tests for _tool_enabled helper."""

    def test_returns_true_for_bool_true_entry(self) -> None:
        config = {"python": {"tools": {"ruff": True}}}
        assert _tool_enabled(config, "python", "ruff", False) is True

    def test_returns_false_for_bool_false_entry(self) -> None:
        config = {"python": {"tools": {"ruff": False}}}
        assert _tool_enabled(config, "python", "ruff", True) is False

    def test_returns_enabled_from_dict_entry(self) -> None:
        config = {"python": {"tools": {"pytest": {"enabled": True}}}}
        assert _tool_enabled(config, "python", "pytest", False) is True

    def test_returns_default_when_enabled_not_bool(self) -> None:
        config = {"python": {"tools": {"pytest": {"enabled": "yes"}}}}
        assert _tool_enabled(config, "python", "pytest", True) is True

    def test_returns_default_for_missing_tool(self) -> None:
        config = {"python": {"tools": {}}}
        assert _tool_enabled(config, "python", "missing", True) is True


class TestToolInt:
    """Tests for _tool_int helper."""

    def test_returns_int_from_tool_config(self) -> None:
        config = {"python": {"tools": {"pytest": {"min_coverage": 85}}}}
        assert _tool_int(config, "python", "pytest", "min_coverage", 70) == 85

    def test_returns_default_for_non_dict_entry(self) -> None:
        config = {"python": {"tools": {"pytest": True}}}
        assert _tool_int(config, "python", "pytest", "min_coverage", 70) == 70

    def test_returns_default_for_missing_key(self) -> None:
        config = {"python": {"tools": {"pytest": {}}}}
        assert _tool_int(config, "python", "pytest", "min_coverage", 70) == 70

    def test_converts_string_to_int(self) -> None:
        config = {"python": {"tools": {"pytest": {"min_coverage": "80"}}}}
        assert _tool_int(config, "python", "pytest", "min_coverage", 70) == 80

    def test_returns_default_for_non_numeric_string(self) -> None:
        config = {"python": {"tools": {"pytest": {"min_coverage": "high"}}}}
        assert _tool_int(config, "python", "pytest", "min_coverage", 70) == 70


class TestCmdConfigOutputs:
    """Tests for cmd_config_outputs command."""

    def test_loads_config_and_returns_outputs_json_mode(self, tmp_path: Path) -> None:
        """Test JSON mode returns outputs dict."""
        config_file = tmp_path / ".ci-hub.yml"
        config_file.write_text("language: python\nrepo:\n  owner: test\n  name: repo\n")

        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python", "python": {"version": "3.11"}}
            result = cmd_config_outputs(args)

        assert result.exit_code == EXIT_SUCCESS
        assert result.data is not None
        assert result.data["outputs"]["language"] == "python"
        assert result.data["outputs"]["python_version"] == "3.11"

    def test_respects_github_summary_config(self, tmp_path: Path) -> None:
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "reports": {"github_summary": {"enabled": False}},
            }
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["write_github_summary"] == "false"

    def test_outputs_to_stdout_when_no_github_output(self, tmp_path: Path) -> None:
        """Test outputs returned in CommandResult for human rendering."""
        args = argparse.Namespace(repo=str(tmp_path), json=False, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_config_outputs(args)

        assert result.exit_code == EXIT_SUCCESS
        assert "language=python" in result.data["raw_output"]

    def test_writes_to_github_output_file(self, tmp_path: Path) -> None:
        """Test writes to GITHUB_OUTPUT when flag set."""
        output_file = tmp_path / "github_output.txt"
        args = argparse.Namespace(repo=str(tmp_path), json=False, workdir=None, github_output=True)

        with (
            patch("cihub.commands.config_outputs.load_ci_config") as mock_load,
            patch.dict(os.environ, {"GITHUB_OUTPUT": str(output_file)}),
        ):
            mock_load.return_value = {"language": "java"}
            result = cmd_config_outputs(args)

        assert result.exit_code == EXIT_SUCCESS
        content = output_file.read_text()
        assert "language=java" in content

    def test_uses_workdir_from_args(self, tmp_path: Path) -> None:
        """Test workdir from args takes precedence."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir="custom/path", github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["workdir"] == "custom/path"

    def test_uses_workdir_from_config(self, tmp_path: Path) -> None:
        """Test workdir from config when not in args."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python", "workdir": "src/app"}
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["workdir"] == "src/app"

    def test_falls_back_to_repo_subdir(self, tmp_path: Path) -> None:
        """Test falls back to repo.subdir when workdir missing."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python", "repo": {"subdir": "backend"}}
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["workdir"] == "backend"

    def test_defaults_workdir_to_dot(self, tmp_path: Path) -> None:
        """Test workdir defaults to '.' when nothing specified."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["workdir"] == "."

    def test_extracts_python_tool_toggles(self, tmp_path: Path) -> None:
        """Test Python tool toggles are extracted."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {
                    "tools": {
                        "pytest": True,
                        "ruff": False,
                        "bandit": {"enabled": True},
                        "mypy": {"enabled": True},
                    }
                },
            }
            result = cmd_config_outputs(args)

        outputs = result.data["outputs"]
        assert outputs["run_pytest"] == "true"
        assert outputs["run_ruff"] == "false"
        assert outputs["run_bandit"] == "true"
        assert outputs["run_mypy"] == "true"

    def test_extracts_java_tool_toggles(self, tmp_path: Path) -> None:
        """Test Java tool toggles are extracted."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "java",
                "java": {
                    "version": "17",
                    "build_tool": "gradle",
                    "tools": {
                        "jacoco": True,
                        "checkstyle": False,
                        "pitest": {"enabled": True},
                    },
                },
            }
            result = cmd_config_outputs(args)

        outputs = result.data["outputs"]
        assert outputs["language"] == "java"
        assert outputs["java_version"] == "17"
        assert outputs["build_tool"] == "gradle"
        assert outputs["run_jacoco"] == "true"
        assert outputs["run_checkstyle"] == "false"
        assert outputs["run_pitest"] == "true"

    def test_extracts_thresholds(self, tmp_path: Path) -> None:
        """Test threshold extraction."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {
                    "tools": {
                        "pytest": {"min_coverage": 85},
                        "mutmut": {"min_mutation_score": 75},
                    }
                },
            }
            result = cmd_config_outputs(args)

        outputs = result.data["outputs"]
        assert outputs["coverage_min"] == "85"
        assert outputs["mutation_score_min"] == "75"

    def test_global_thresholds_override_tool_thresholds(self, tmp_path: Path) -> None:
        """Test global thresholds override language-specific ones."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"pytest": {"min_coverage": 60}}},
                "thresholds": {"coverage_min": 90},
            }
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["coverage_min"] == "90"

    def test_java_language_uses_java_thresholds(self, tmp_path: Path) -> None:
        """Test Java language uses Java-specific thresholds."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "java",
                "java": {
                    "tools": {
                        "jacoco": {"min_coverage": 80},
                        "pitest": {"min_mutation_score": 65},
                    }
                },
            }
            result = cmd_config_outputs(args)

        outputs = result.data["outputs"]
        assert outputs["coverage_min"] == "80"
        assert outputs["mutation_score_min"] == "65"

    def test_returns_failure_on_config_load_error_json_mode(self, tmp_path: Path) -> None:
        """Test returns failure when config fails to load (JSON mode)."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.side_effect = Exception("Config not found")
            result = cmd_config_outputs(args)

        assert result.exit_code == EXIT_FAILURE
        assert "Failed to load config" in result.summary
        assert len(result.problems) == 1

    def test_returns_failure_on_config_load_error_text_mode(self, tmp_path: Path) -> None:
        """Test returns failure when config fails to load (text mode)."""
        args = argparse.Namespace(repo=str(tmp_path), json=False, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.side_effect = Exception("Config not found")
            result = cmd_config_outputs(args)

        assert result.exit_code == EXIT_FAILURE
        assert "Failed to load config" in result.summary

    def test_uses_language_from_repo_section(self, tmp_path: Path) -> None:
        """Test language falls back to repo.language."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {"repo": {"language": "java"}}
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["language"] == "java"

    def test_nvd_api_key_toggle(self, tmp_path: Path) -> None:
        """Test NVD API key toggle logic."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        # When owasp enabled and use_nvd_api_key not explicitly false
        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "java",
                "java": {"tools": {"owasp": {"enabled": True}}},
            }
            result = cmd_config_outputs(args)
        assert result.data["outputs"]["use_nvd_api_key"] == "true"

        # When owasp disabled
        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "java",
                "java": {"tools": {"owasp": False}},
            }
            result = cmd_config_outputs(args)
        assert result.data["outputs"]["use_nvd_api_key"] == "false"

    def test_vuln_thresholds(self, tmp_path: Path) -> None:
        """Test max_critical_vulns and max_high_vulns extraction."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "thresholds": {"max_critical_vulns": 5, "max_high_vulns": 10},
            }
            result = cmd_config_outputs(args)

        outputs = result.data["outputs"]
        assert outputs["max_critical_vulns"] == "5"
        assert outputs["max_high_vulns"] == "10"

    def test_retention_days(self, tmp_path: Path) -> None:
        """Test retention_days extraction."""
        args = argparse.Namespace(repo=str(tmp_path), json=True, workdir=None, github_output=False)

        with patch("cihub.commands.config_outputs.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "reports": {"retention_days": 14},
            }
            result = cmd_config_outputs(args)

        assert result.data["outputs"]["retention_days"] == "14"



"""CI execution engine for the services layer.

This package splits the ci_engine module into logical submodules while
maintaining backward compatibility by re-exporting all public symbols.
"""

from __future__ import annotations

import json
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Mapping

from cihub.ci_config import load_ci_config, load_hub_config
from cihub.ci_runner import run_java_build  # Keep for backward compat re-export
from cihub.core.languages import get_strategy
from cihub.exit_codes import EXIT_FAILURE, EXIT_INTERNAL_ERROR, EXIT_SUCCESS
from cihub.reporting import render_summary
from cihub.services.types import RunCIOptions, ServiceResult
from cihub.tools.registry import (
    JAVA_TOOLS,
    PYTHON_TOOLS,
    RESERVED_FEATURES,
    get_all_tools_from_config,
    get_runners,
    is_custom_tool,
)
from cihub.utils import (
    detect_java_project_type,
    get_git_branch,
    get_git_remote,
    get_repo_name,
    parse_repo_from_remote,
    resolve_executable,
    validate_subdir,
)

# Backward compatibility aliases (deprecated, use non-underscore versions)
_detect_java_project_type = detect_java_project_type
_get_repo_name = get_repo_name

# Import from submodules
from .gates import (
    _build_context,
    _check_require_run_or_fail,
    _evaluate_java_gates,
    _evaluate_python_gates,
    _tool_requires_run_or_fail,
)
from .helpers import (
    _apply_env_overrides,
    _apply_force_all_tools,
    _get_env_name,
    _get_env_value,
    _get_git_commit,
    _parse_env_bool,
    _resolve_workdir,
    _set_tool_enabled,
    _split_problems,
    _tool_enabled,
    _tool_gate_enabled,
    _warn_reserved_features,
)
from .io import _collect_codecov_files, _run_codecov_upload
from .java_tools import _run_java_tools
from .notifications import _notify, _send_email, _send_slack
from .python_tools import (
    _install_python_dependencies,
    _run_dep_command,
    _run_python_tools,
)
from .validation import _self_validate_report


@dataclass
class CiRunResult(ServiceResult):
    """Result of running cihub ci via the services layer."""

    exit_code: int = 0
    report_path: Path | None = None
    summary_path: Path | None = None
    report: dict[str, Any] = field(default_factory=dict)
    summary_text: str = ""
    artifacts: dict[str, str] = field(default_factory=dict)
    data: dict[str, Any] = field(default_factory=dict)
    problems: list[dict[str, Any]] = field(default_factory=list)


# Tool runner dictionaries (backward compatibility)
# New code should use: from cihub.tools.registry import get_runners
# These are lazily evaluated on first access.

# Mapping of deprecated run_* function names to (language, tool_name)
_RUNNER_COMPAT_MAP: dict[str, tuple[str, str]] = {
    "run_pytest": ("python", "pytest"),
    "run_ruff": ("python", "ruff"),
    "run_black": ("python", "black"),
    "run_isort": ("python", "isort"),
    "run_mypy": ("python", "mypy"),
    "run_bandit": ("python", "bandit"),
    "run_pip_audit": ("python", "pip_audit"),
    "run_mutmut": ("python", "mutmut"),
    "run_sbom": ("python", "sbom"),
    "run_semgrep": ("python", "semgrep"),
    "run_trivy": ("python", "trivy"),
    "run_docker": ("python", "docker"),
    "run_jacoco": ("java", "jacoco"),
    "run_pitest": ("java", "pitest"),
    "run_checkstyle": ("java", "checkstyle"),
    "run_spotbugs": ("java", "spotbugs"),
    "run_pmd": ("java", "pmd"),
    "run_owasp": ("java", "owasp"),
}


def __getattr__(name: str) -> Any:
    """Lazy module attribute access for backward compatibility.

    Provides backward-compatible access to:
    - PYTHON_RUNNERS, JAVA_RUNNERS dicts
    - Individual run_* functions (run_pytest, run_ruff, etc.)
    """
    if name == "PYTHON_RUNNERS":
        return dict(get_runners("python"))  # Return copy for isolation
    if name == "JAVA_RUNNERS":
        return dict(get_runners("java"))  # Return copy for isolation
    if name in _RUNNER_COMPAT_MAP:
        from cihub.tools.registry import get_runner

        lang, tool = _RUNNER_COMPAT_MAP[name]
        runner = get_runner(tool, lang)
        if runner is None:
            raise AttributeError(
                f"Runner '{name}' not found in registry for {lang}/{tool}. "
                f"Check cihub/tools/registry.py _load_{lang}_runners()."
            )
        return runner
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def run_ci(
    repo_path: Path,
    *,
    options: RunCIOptions | None = None,
    output_dir: Path | None = None,
    report_path: Path | None = None,
    summary_path: Path | None = None,
    workdir: str | None = None,
    install_deps: bool = False,
    correlation_id: str | None = None,
    no_summary: bool = False,
    write_github_summary: bool | None = None,
    config_from_hub: str | None = None,
    env: Mapping[str, str] | None = None,
) -> CiRunResult:
    """Run CI pipeline for a repository.

    Args:
        repo_path: Path to the repository root
        options: RunCIOptions object (preferred). If provided, individual kwargs are ignored.
        output_dir: Directory for output files (default: .cihub)
        report_path: Path for report.json (default: output_dir/report.json)
        summary_path: Path for summary.md (default: output_dir/summary.md)
        workdir: Subdirectory to run CI in
        install_deps: Whether to install Python dependencies
        correlation_id: Correlation ID for tracing
        no_summary: Skip writing summary file
        write_github_summary: Write to GITHUB_STEP_SUMMARY
        config_from_hub: Load config from hub config files
        env: Environment variable mapping (default: os.environ)

    Returns:
        CiRunResult with exit code, report, and any problems

    Example:
        # Using options (preferred)
        opts = RunCIOptions(install_deps=True)
        result = run_ci(repo_path, options=opts)

        # Using kwargs (backward compatible)
        result = run_ci(repo_path, install_deps=True)
    """
    # If options provided, extract values from it; otherwise use kwargs
    if options is not None:
        output_dir = options.output_dir
        report_path = options.report_path
        summary_path = options.summary_path
        workdir = options.workdir
        install_deps = options.install_deps
        correlation_id = options.correlation_id
        no_summary = options.no_summary
        write_github_summary = options.write_github_summary
        config_from_hub = options.config_from_hub
        env = options.env

    repo_path = repo_path.resolve()
    output_dir = Path(output_dir or ".cihub")
    if not output_dir.is_absolute():
        output_dir = repo_path / output_dir
    output_dir = output_dir.resolve()
    output_dir.mkdir(parents=True, exist_ok=True)

    env_map = dict(env) if env is not None else dict(os.environ)

    try:
        if config_from_hub:
            config = load_hub_config(config_from_hub, repo_path)
        else:
            config = load_ci_config(repo_path)
    except Exception as exc:
        message = f"Failed to load config: {exc}"
        config_problems = [{"severity": "error", "message": message, "code": "CIHUB-CI-CONFIG"}]
        errors, warnings = _split_problems(config_problems)
        return CiRunResult(
            success=False,
            errors=errors,
            warnings=warnings,
            exit_code=EXIT_FAILURE,
            problems=config_problems,
        )

    language = config.get("language") or ""
    run_workdir = _resolve_workdir(repo_path, config, workdir)
    problems: list[dict[str, Any]] = []
    _apply_force_all_tools(config, language)
    _apply_env_overrides(config, language, env_map, problems)
    _warn_reserved_features(config, problems)

    # -------------------------------------------------------------------------
    # Language Strategy Pattern: get strategy early, fail fast for unsupported
    # -------------------------------------------------------------------------
    try:
        strategy = get_strategy(language)
    except ValueError:
        message = f"cihub ci supports python or java (got '{language}')"
        problems = [{"severity": "error", "message": message, "code": "CIHUB-CI-LANGUAGE"}]
        errors, warnings = _split_problems(problems)
        return CiRunResult(
            success=False,
            errors=errors,
            warnings=warnings,
            exit_code=EXIT_FAILURE,
            problems=problems,
        )

    report: dict[str, Any] = {}
    tool_outputs: dict[str, dict[str, Any]] = {}
    tools_ran: dict[str, bool] = {}
    tools_success: dict[str, bool] = {}
    gate_failures: list[str] = []

    # -------------------------------------------------------------------------
    # Run tools via strategy (language-specific kwargs merged via strategy)
    # -------------------------------------------------------------------------
    try:
        # Get language-specific kwargs merged with caller-provided kwargs
        run_kwargs = strategy.get_run_kwargs(config, install_deps=install_deps)
        tool_outputs, tools_ran, tools_success = strategy.run_tools(
            config,
            repo_path,
            run_workdir,
            output_dir,
            problems,
            **run_kwargs,
        )
    except Exception as exc:
        message = f"Tool execution failed: {exc}"
        problems.append(
            {
                "severity": "error",
                "message": message,
                "code": "CIHUB-CI-TOOL-FAILURE",
            }
        )
        errors, warnings = _split_problems(problems)
        return CiRunResult(
            success=False,
            errors=errors,
            warnings=warnings,
            exit_code=EXIT_INTERNAL_ERROR,
            problems=problems,
        )

    # -------------------------------------------------------------------------
    # Build tools_configured and tools_require_run using strategy
    # Include custom tools (x-* prefix) from config alongside built-in tools
    # -------------------------------------------------------------------------
    all_tools = get_all_tools_from_config(config, language)
    # Custom tools (x-* prefix) default to enabled=True per schema; built-in tools default to False
    tools_configured = {tool: _tool_enabled(config, tool, language, default=is_custom_tool(tool)) for tool in all_tools}
    tools_require_run = {tool: _tool_requires_run_or_fail(tool, config, language) for tool in all_tools}

    # -------------------------------------------------------------------------
    # Resolve thresholds via strategy
    # -------------------------------------------------------------------------
    thresholds = strategy.resolve_thresholds(config)

    # -------------------------------------------------------------------------
    # Extract docker config via strategy (encapsulates language-specific access)
    # -------------------------------------------------------------------------
    docker_cfg = strategy.get_docker_config(config)
    docker_compose = docker_cfg.get("compose_file")
    docker_health = docker_cfg.get("health_endpoint")
    # Apply language-specific default for compose_file
    if tools_configured.get("docker") and not docker_compose:
        docker_compose = strategy.get_docker_compose_default()
    # Only include docker config if docker is actually configured
    if not tools_configured.get("docker"):
        docker_compose = None
        docker_health = None

    # -------------------------------------------------------------------------
    # Build context (language-specific extras via strategy)
    # -------------------------------------------------------------------------
    context_kwargs: dict[str, Any] = {
        "docker_compose_file": docker_compose,
        "docker_health_endpoint": docker_health,
    }
    # Add language-specific context fields (e.g., Java's build_tool, project_type)
    context_extras = strategy.get_context_extras(config, repo_path / run_workdir)
    context_kwargs.update(context_extras)

    context = _build_context(repo_path, config, run_workdir, correlation_id, **context_kwargs)

    # -------------------------------------------------------------------------
    # Build report and evaluate gates via strategy
    # -------------------------------------------------------------------------
    report = strategy.build_report(
        config,
        tool_outputs,
        tools_configured,
        tools_ran,
        tools_success,
        thresholds,
        context,
        tools_require_run=tools_require_run,
    )
    gate_failures = strategy.evaluate_gates(report, thresholds, tools_configured, config)

    resolved_report_path = report_path or output_dir / "report.json"
    if not resolved_report_path.is_absolute():
        resolved_report_path = repo_path / resolved_report_path
    resolved_report_path.write_text(json.dumps(report, indent=2), encoding="utf-8")

    github_summary_cfg = config.get("reports", {}).get("github_summary", {}) or {}
    include_metrics = bool(github_summary_cfg.get("include_metrics", True))
    summary_text = render_summary(report, include_metrics=include_metrics)
    if write_github_summary is None:
        write_summary = bool(github_summary_cfg.get("enabled", True))
    else:
        write_summary = bool(write_github_summary)

    resolved_summary_path: Path | None = None
    if not no_summary:
        resolved_summary_path = summary_path or output_dir / "summary.md"
        if not resolved_summary_path.is_absolute():
            resolved_summary_path = repo_path / resolved_summary_path
        resolved_summary_path.write_text(summary_text, encoding="utf-8")

    github_summary_env = env_map.get("GITHUB_STEP_SUMMARY")
    if write_summary and github_summary_env:
        Path(github_summary_env).write_text(summary_text, encoding="utf-8")

    # Self-validation: catch contradictions between report and summary
    _self_validate_report(report, summary_text, output_dir, problems, env_map)

    codecov_cfg = config.get("reports", {}).get("codecov", {}) or {}
    if codecov_cfg.get("enabled", True):
        files = _collect_codecov_files(language, output_dir, tool_outputs)
        _run_codecov_upload(
            files,
            bool(codecov_cfg.get("fail_ci_on_error", False)),
            problems,
        )

    if gate_failures:
        problems.extend(
            [
                {
                    "severity": "error",
                    "message": failure,
                    "code": "CIHUB-CI-GATE",
                }
                for failure in gate_failures
            ]
        )

    has_errors = any(p.get("severity") == "error" for p in problems)
    _notify(not has_errors, config, report, problems, env_map)
    exit_code = EXIT_FAILURE if has_errors else EXIT_SUCCESS
    errors, warnings = _split_problems(problems)

    artifacts: dict[str, str] = {"report": str(resolved_report_path)}
    data: dict[str, str] = {"report_path": str(resolved_report_path)}
    if resolved_summary_path:
        artifacts["summary"] = str(resolved_summary_path)
        data["summary_path"] = str(resolved_summary_path)

    return CiRunResult(
        success=exit_code == EXIT_SUCCESS,
        errors=errors,
        warnings=warnings,
        exit_code=exit_code,
        report_path=resolved_report_path,
        summary_path=resolved_summary_path,
        report=report,
        summary_text=summary_text,
        artifacts=artifacts,
        data=data,
        problems=problems,
    )


# ============================================================================
# Public API - Re-exports for backward compatibility
# ============================================================================

__all__ = [
    # Main entry point
    "run_ci",
    "CiRunResult",
    "RunCIOptions",
    # Runner dictionaries
    "PYTHON_RUNNERS",
    "JAVA_RUNNERS",
    # Helpers from helpers.py
    "get_repo_name",
    "_get_git_commit",
    "_resolve_workdir",
    "detect_java_project_type",
    # Backward compatibility aliases (deprecated)
    "_get_repo_name",
    "_detect_java_project_type",
    "_tool_enabled",
    "_tool_gate_enabled",
    "_parse_env_bool",
    "_get_env_name",
    "_get_env_value",
    "_warn_reserved_features",
    "_set_tool_enabled",
    "_apply_env_overrides",
    "_apply_force_all_tools",
    "_split_problems",
    # Helpers from io.py
    "_collect_codecov_files",
    "_run_codecov_upload",
    # Helpers from notifications.py
    "_send_slack",
    "_send_email",
    "_notify",
    # Helpers from python_tools.py
    "_run_dep_command",
    "_install_python_dependencies",
    "_run_python_tools",
    # Helpers from java_tools.py
    "_run_java_tools",
    # Helpers from gates.py
    "_build_context",
    "_evaluate_python_gates",
    "_evaluate_java_gates",
    "_tool_requires_run_or_fail",
    "_check_require_run_or_fail",
    # Helpers from validation.py
    "_self_validate_report",
    # Re-exports from cihub.cli for backward compatibility
    "get_git_branch",
    "get_git_remote",
    "parse_repo_from_remote",
    "resolve_executable",
    "validate_subdir",
    # Re-exports from cihub.tools.registry
    "PYTHON_TOOLS",
    "JAVA_TOOLS",
    "RESERVED_FEATURES",
    # Re-exports from cihub.ci_runner for backward compatibility
    "run_java_build",
]



"""Tests for discover command handler."""

from __future__ import annotations

import argparse
import json
from pathlib import Path
from unittest import mock

import pytest

from cihub.commands import discover as discover_cmd
from cihub.exit_codes import EXIT_FAILURE, EXIT_SUCCESS
from cihub.services import DiscoveryResult
from cihub.services.types import RepoEntry


def test_discover_conflicting_filters() -> None:
    args = argparse.Namespace(
        hub_root=None,
        json=False,
        run_group="",
        repos="",
        central_only=True,
        dispatch_only=True,
        github_output=False,
    )

    result = discover_cmd.cmd_discover(args)

    assert result.exit_code == EXIT_FAILURE
    assert "Choose only one" in result.summary


def test_discover_json_with_github_output(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    entry = RepoEntry(
        config_basename="repo1",
        name="repo1",
        owner="owner",
        language="python",
        branch="main",
        subdir="",
        subdir_safe="",
        run_group="full",
        dispatch_enabled=True,
        dispatch_workflow="hub-ci.yml",
        use_central_runner=True,
        tools={"run_pytest": True},
        thresholds={},
        java_version=None,
        python_version="3.12",
        build_tool=None,
        retention_days=30,
        write_github_summary=True,
    )
    result_obj = DiscoveryResult(success=True, entries=[entry], warnings=[])

    args = argparse.Namespace(
        hub_root=str(tmp_path),
        json=True,
        run_group="",
        repos="",
        central_only=False,
        dispatch_only=False,
        github_output=True,
    )

    output_file = tmp_path / "github_output"
    monkeypatch.setenv("GITHUB_OUTPUT", str(output_file))

    with mock.patch.object(discover_cmd, "discover_repositories", return_value=result_obj):
        result = discover_cmd.cmd_discover(args)

    assert result.exit_code == EXIT_SUCCESS
    assert result.data["count"] == 1
    output_text = output_file.read_text(encoding="utf-8")
    assert "matrix=" in output_text
    assert "count=1" in output_text
    payload = json.loads(output_text.split("matrix=", 1)[1].splitlines()[0])
    assert payload["include"][0]["name"] == "repo1"


def test_discover_no_entries_json(tmp_path: Path) -> None:
    result_obj = DiscoveryResult(success=True, entries=[], warnings=[])
    args = argparse.Namespace(
        hub_root=str(tmp_path),
        json=True,
        run_group="",
        repos="",
        central_only=False,
        dispatch_only=False,
        github_output=False,
    )

    with mock.patch.object(discover_cmd, "discover_repositories", return_value=result_obj):
        result = discover_cmd.cmd_discover(args)

    assert result.exit_code == EXIT_FAILURE
    assert "No repositories found" in result.summary



"""Debug helpers for CLI diagnostics."""

from __future__ import annotations

import sys
from typing import Iterable, Mapping, TextIO

from cihub.utils.env import env_bool


def debug_context_enabled(env: Mapping[str, str] | None = None) -> bool:
    """Return True if CIHUB_DEBUG_CONTEXT is enabled."""
    return env_bool("CIHUB_DEBUG_CONTEXT", default=False, env=env)


def emit_debug_context(
    title: str,
    entries: Iterable[tuple[str, str | None]],
    *,
    env: Mapping[str, str] | None = None,
    file: TextIO | None = None,
) -> None:
    """Emit a compact debug context block to stderr when enabled."""
    if not debug_context_enabled(env=env):
        return
    output = file or sys.stderr
    lines = [f"[cihub debug] {title}"]
    for key, value in entries:
        if value is None:
            continue
        text = str(value)
        if not text:
            continue
        lines.append(f"  {key}: {text}")
    output.write("\n".join(lines) + "\n")



"""Tests for cihub.commands.pom module (fix-pom and fix-deps commands)."""

from __future__ import annotations

import argparse
import sys
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))

from cihub.commands.pom import cmd_fix_deps, cmd_fix_pom  # noqa: E402
from cihub.exit_codes import EXIT_FAILURE, EXIT_SUCCESS, EXIT_USAGE  # noqa: E402
from cihub.types import CommandResult  # noqa: E402


def write_ci_hub_config(path: Path, language: str = "java", build_tool: str = "maven") -> None:
    """Helper to create .ci-hub.yml config file."""
    config = f"""
repo:
  owner: test
  name: example
  default_branch: main
  language: {language}
language: {language}
{language}:
  version: '21'
  build_tool: {build_tool}
  tools:
    jacoco:
      enabled: true
    checkstyle:
      enabled: true
    jqwik:
      enabled: true
"""
    (path / ".ci-hub.yml").write_text(config, encoding="utf-8")


def write_pom(path: Path, body: str) -> None:
    """Helper to write pom.xml file."""
    path.write_text(body, encoding="utf-8")


# ==============================================================================
# Tests for cmd_fix_pom
# ==============================================================================


class TestCmdFixPom:
    """Tests for the fix-pom command."""

    def test_fix_pom_missing_config(self, tmp_path: Path) -> None:
        """fix-pom returns error when .ci-hub.yml is missing."""
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_USAGE
        assert "Config not found" in result.summary

    def test_fix_pom_non_java_repo(self, tmp_path: Path) -> None:
        """fix-pom exits gracefully for non-Java repos."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: python
python:
  version: '3.12'
  tools:
    pytest:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        assert "fix-pom is only supported for Java repos" in result.summary

    def test_fix_pom_gradle_repo(self, tmp_path: Path) -> None:
        """fix-pom exits gracefully for Gradle repos."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: gradle
  tools:
    jacoco:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        assert "fix-pom only supports Maven repos" in result.summary

    def test_fix_pom_dry_run_shows_diff(self, tmp_path: Path) -> None:
        """fix-pom in dry-run mode shows diff without modifying file."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )
        original_content = pom_path.read_text(encoding="utf-8")

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        # File should NOT be modified in dry-run
        assert pom_path.read_text(encoding="utf-8") == original_content

        # Should show diff output in raw_output
        raw_output = result.data.get("raw_output", "")
        assert "---" in raw_output or "+++" in raw_output or "jacoco" in raw_output.lower()

    def test_fix_pom_apply_modifies_file(self, tmp_path: Path) -> None:
        """fix-pom with --apply modifies pom.xml."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        # File SHOULD be modified
        updated_content = pom_path.read_text(encoding="utf-8")
        assert "<artifactId>jacoco-maven-plugin</artifactId>" in updated_content

        items = result.data.get("items", [])
        assert any("pom.xml updated" in item for item in items)

    def test_fix_pom_no_changes_needed(self, tmp_path: Path) -> None:
        """fix-pom reports no changes when plugins already present."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: maven
  tools:
    jacoco:
      enabled: false
    checkstyle:
      enabled: false
    spotbugs:
      enabled: false
    pmd:
      enabled: false
    owasp:
      enabled: false
    pitest:
      enabled: false
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        items = result.data.get("items", [])
        assert any("No pom.xml changes needed" in item for item in items)

    def test_fix_pom_missing_pom_file(self, tmp_path: Path) -> None:
        """fix-pom returns error when pom.xml is missing."""
        write_ci_hub_config(tmp_path)
        # No pom.xml created

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_FAILURE
        assert any("pom.xml not found" in p.get("message", "") for p in result.problems)


# ==============================================================================
# Tests for cmd_fix_deps
# ==============================================================================


class TestCmdFixDeps:
    """Tests for the fix-deps command."""

    def test_fix_deps_missing_config(self, tmp_path: Path) -> None:
        """fix-deps returns error when .ci-hub.yml is missing."""
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_USAGE
        assert "Config not found" in result.summary

    def test_fix_deps_non_java_repo(self, tmp_path: Path) -> None:
        """fix-deps exits gracefully for non-Java repos."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: python
python:
  version: '3.12'
  tools:
    pytest:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        assert "fix-deps is only supported for Java repos" in result.summary

    def test_fix_deps_gradle_repo(self, tmp_path: Path) -> None:
        """fix-deps exits gracefully for Gradle repos."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: gradle
  tools:
    jqwik:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        assert "fix-deps only supports Maven repos" in result.summary

    def test_fix_deps_dry_run_shows_diff(self, tmp_path: Path) -> None:
        """fix-deps in dry-run mode shows diff without modifying file."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
  <dependencies>
    <dependency>
      <groupId>org.junit.jupiter</groupId>
      <artifactId>junit-jupiter</artifactId>
      <version>5.10.0</version>
      <scope>test</scope>
    </dependency>
  </dependencies>
</project>
""",
        )
        original_content = pom_path.read_text(encoding="utf-8")

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        # File should NOT be modified in dry-run
        assert pom_path.read_text(encoding="utf-8") == original_content

        # Should show diff output or warning about jqwik in raw_output or items
        raw_output = result.data.get("raw_output", "")
        items = result.data.get("items", [])
        combined = raw_output + " ".join(items)
        assert "---" in combined or "jqwik" in combined.lower()

    def test_fix_deps_apply_modifies_file(self, tmp_path: Path) -> None:
        """fix-deps with --apply modifies pom.xml."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
  <dependencies>
    <dependency>
      <groupId>org.junit.jupiter</groupId>
      <artifactId>junit-jupiter</artifactId>
      <version>5.10.0</version>
      <scope>test</scope>
    </dependency>
  </dependencies>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        # File SHOULD be modified
        updated_content = pom_path.read_text(encoding="utf-8")
        assert "<artifactId>jqwik</artifactId>" in updated_content

        items = result.data.get("items", [])
        assert any("updated" in item for item in items)

    def test_fix_deps_no_changes_needed(self, tmp_path: Path) -> None:
        """fix-deps reports no changes when dependencies already present."""
        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: maven
  tools:
    jqwik:
      enabled: false
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS
        items = result.data.get("items", [])
        assert any("No dependency changes needed" in item for item in items)


# ==============================================================================
# Edge case tests
# ==============================================================================


class TestPomCommandsEdgeCases:
    """Edge case tests for pom commands."""

    def test_fix_pom_with_subdir(self, tmp_path: Path) -> None:
        """fix-pom handles subdir config correctly."""
        subdir = "services/backend"
        (tmp_path / subdir).mkdir(parents=True)

        config = f"""
repo:
  owner: test
  name: example
  default_branch: main
  language: java
  subdir: {subdir}
java:
  version: '21'
  build_tool: maven
  tools:
    jacoco:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")

        pom_path = tmp_path / subdir / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        updated_content = pom_path.read_text(encoding="utf-8")
        assert "<artifactId>jacoco-maven-plugin</artifactId>" in updated_content

    def test_fix_pom_invalid_xml(self, tmp_path: Path) -> None:
        """fix-pom handles invalid XML gracefully."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(pom_path, "<project>invalid xml without closing tag")

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        # Invalid XML should trigger a warning or non-zero exit
        items = result.data.get("items", [])
        combined = " ".join(items)
        assert result.exit_code != EXIT_SUCCESS or "Invalid pom.xml" in combined or "warning" in combined.lower()

    def test_fix_deps_multi_module_project(self, tmp_path: Path) -> None:
        """fix-deps handles multi-module Maven projects."""
        write_ci_hub_config(tmp_path)

        # Create parent pom with modules
        parent_pom = tmp_path / "pom.xml"
        write_pom(
            parent_pom,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>parent</artifactId>
  <version>1.0.0</version>
  <packaging>pom</packaging>
  <modules>
    <module>module-a</module>
  </modules>
</project>
""",
        )

        # Create module pom
        module_dir = tmp_path / "module-a"
        module_dir.mkdir()
        module_pom = module_dir / "pom.xml"
        write_pom(
            module_pom,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <parent>
    <groupId>com.example</groupId>
    <artifactId>parent</artifactId>
    <version>1.0.0</version>
  </parent>
  <artifactId>module-a</artifactId>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

        # Module pom should be updated with jqwik dependency
        updated_content = module_pom.read_text(encoding="utf-8")
        assert "<artifactId>jqwik</artifactId>" in updated_content

    def test_fix_pom_resolves_repo_path(self, tmp_path: Path) -> None:
        """fix-pom resolves relative paths correctly."""
        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        # Use the actual path (not relative)
        args = argparse.Namespace(repo=str(tmp_path), apply=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == EXIT_SUCCESS

    def test_fix_deps_missing_pom_no_error_for_missing_file(self, tmp_path: Path) -> None:
        """fix-deps returns no error when pom.xml is missing (unlike fix-pom)."""
        write_ci_hub_config(tmp_path)
        # No pom.xml created

        args = argparse.Namespace(repo=str(tmp_path), apply=False)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        # fix-deps just reports no changes needed when pom missing
        assert result.exit_code == EXIT_SUCCESS
        items = result.data.get("items", [])
        assert any("No dependency changes needed" in item for item in items)


# ==============================================================================
# JSON Mode Tests
# ==============================================================================


class TestPomCommandsJsonMode:
    """JSON mode tests for pom commands."""

    def test_fix_pom_missing_config_json_mode(self, tmp_path: Path) -> None:
        """fix-pom returns CommandResult for missing config in JSON mode."""
        from cihub.types import CommandResult

        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 2
        assert "Config not found" in result.summary

    def test_fix_pom_non_java_json_mode(self, tmp_path: Path) -> None:
        """fix-pom returns CommandResult for non-Java repo in JSON mode."""
        from cihub.types import CommandResult

        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: python
python:
  version: '3.12'
  tools:
    pytest:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert "Java" in result.summary

    def test_fix_pom_gradle_json_mode(self, tmp_path: Path) -> None:
        """fix-pom returns CommandResult for Gradle repo in JSON mode."""
        from cihub.types import CommandResult

        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: gradle
  tools:
    jacoco:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert "Maven" in result.summary

    def test_fix_pom_apply_json_mode(self, tmp_path: Path) -> None:
        """fix-pom with apply in JSON mode returns CommandResult."""
        from cihub.types import CommandResult

        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True, json=True)
        result = cmd_fix_pom(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert result.data["applied"] is True

    def test_fix_deps_missing_config_json_mode(self, tmp_path: Path) -> None:
        """fix-deps returns CommandResult for missing config in JSON mode."""
        from cihub.types import CommandResult

        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 2
        assert "Config not found" in result.summary

    def test_fix_deps_non_java_json_mode(self, tmp_path: Path) -> None:
        """fix-deps returns CommandResult for non-Java repo in JSON mode."""
        from cihub.types import CommandResult

        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: python
python:
  version: '3.12'
  tools:
    pytest:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert "Java" in result.summary

    def test_fix_deps_gradle_json_mode(self, tmp_path: Path) -> None:
        """fix-deps returns CommandResult for Gradle repo in JSON mode."""
        from cihub.types import CommandResult

        config = """
repo:
  owner: test
  name: example
  default_branch: main
  language: java
java:
  version: '21'
  build_tool: gradle
  tools:
    jqwik:
      enabled: true
"""
        (tmp_path / ".ci-hub.yml").write_text(config, encoding="utf-8")
        args = argparse.Namespace(repo=str(tmp_path), apply=False, json=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert "Maven" in result.summary

    def test_fix_deps_apply_json_mode(self, tmp_path: Path) -> None:
        """fix-deps with apply in JSON mode returns CommandResult."""
        from cihub.types import CommandResult

        write_ci_hub_config(tmp_path)
        pom_path = tmp_path / "pom.xml"
        write_pom(
            pom_path,
            """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.example</groupId>
  <artifactId>example</artifactId>
  <version>1.0.0</version>
  <dependencies>
    <dependency>
      <groupId>org.junit.jupiter</groupId>
      <artifactId>junit-jupiter</artifactId>
      <version>5.10.0</version>
      <scope>test</scope>
    </dependency>
  </dependencies>
</project>
""",
        )

        args = argparse.Namespace(repo=str(tmp_path), apply=True, json=True)
        result = cmd_fix_deps(args)
        assert isinstance(result, CommandResult)
        assert result.exit_code == 0
        assert result.data["applied"] is True



"""Tests for cihub.commands.run module."""

from __future__ import annotations

import argparse
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

from cihub.ci_runner import ToolResult
from cihub.commands.run import RUNNERS, _tool_enabled, cmd_run
from cihub.exit_codes import EXIT_FAILURE, EXIT_SUCCESS, EXIT_USAGE


class TestToolEnabled:
    """Tests for _tool_enabled helper."""

    def test_returns_true_for_bool_true(self) -> None:
        config = {"python": {"tools": {"ruff": True}}}
        assert _tool_enabled(config, "ruff") is True

    def test_returns_false_for_bool_false(self) -> None:
        config = {"python": {"tools": {"ruff": False}}}
        assert _tool_enabled(config, "ruff") is False

    def test_returns_enabled_from_dict(self) -> None:
        config = {"python": {"tools": {"pytest": {"enabled": True}}}}
        assert _tool_enabled(config, "pytest") is True

    def test_returns_false_when_enabled_is_false(self) -> None:
        config = {"python": {"tools": {"pytest": {"enabled": False}}}}
        assert _tool_enabled(config, "pytest") is False

    def test_returns_false_for_missing_tool(self) -> None:
        config = {"python": {"tools": {}}}
        assert _tool_enabled(config, "missing") is False

    def test_returns_false_for_non_dict_python_block(self) -> None:
        config: dict[str, Any] = {"python": "not_a_dict"}
        assert _tool_enabled(config, "ruff") is False

    def test_returns_false_for_non_dict_tools(self) -> None:
        config = {"python": {"tools": "not_a_dict"}}
        assert _tool_enabled(config, "ruff") is False


class TestRunners:
    """Tests for RUNNERS registry."""

    def test_contains_all_supported_tools(self) -> None:
        expected = {
            "pytest",
            "ruff",
            "black",
            "isort",
            "mypy",
            "bandit",
            "pip_audit",
            "pip-audit",
            "mutmut",
            "semgrep",
            "trivy",
        }
        assert set(RUNNERS.keys()) == expected

    def test_pip_audit_has_alias(self) -> None:
        assert RUNNERS["pip_audit"] is RUNNERS["pip-audit"]


class TestCmdRun:
    """Tests for cmd_run command."""

    def test_returns_usage_error_for_unsupported_tool(self, tmp_path: Path) -> None:
        """Test unsupported tool returns EXIT_USAGE."""
        (tmp_path / ".ci-hub.yml").write_text("language: python\nrepo:\n  owner: test\n  name: repo\n")
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="unsupported",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        assert result.exit_code == EXIT_USAGE
        assert "Unsupported tool" in result.summary

    def test_skips_disabled_tool(self, tmp_path: Path) -> None:
        """Test skips disabled tool without --force."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"ruff": False}},
            }
            result = cmd_run(args)

        assert result.exit_code == EXIT_SUCCESS
        assert "skipped" in result.summary

    def test_runs_tool_when_enabled(self, tmp_path: Path) -> None:
        """Test runs tool when enabled."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "ruff", "ran": True}

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"ruff": MagicMock(return_value=mock_result)}),
        ):
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"ruff": True}},
            }
            result = cmd_run(args)

        assert result.exit_code == EXIT_SUCCESS
        assert "passed" in result.summary

    def test_runs_tool_with_force_flag(self, tmp_path: Path) -> None:
        """Test --force runs tool even when disabled."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="bandit",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,  # Force run
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "bandit"}

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"bandit": MagicMock(return_value=mock_result)}),
        ):
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"bandit": False}},  # Disabled
            }
            result = cmd_run(args)

        assert result.exit_code == EXIT_SUCCESS

    def test_returns_failure_when_tool_fails(self, tmp_path: Path) -> None:
        """Test returns failure when tool fails."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="pytest",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = False
        mock_result.to_payload.return_value = {"tool": "pytest", "ran": True}

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"pytest": MagicMock(return_value=mock_result)}),
        ):
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        assert result.exit_code == EXIT_FAILURE
        assert "failed" in result.summary

    def test_returns_failure_on_config_load_error(self, tmp_path: Path) -> None:
        """Test returns failure when config fails to load."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.side_effect = Exception("Config error")
            result = cmd_run(args)

        assert result.exit_code == EXIT_FAILURE
        assert "Failed to load config" in result.summary

    def test_returns_failure_for_missing_workdir(self, tmp_path: Path) -> None:
        """Test returns failure when workdir doesn't exist."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir="nonexistent",
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        assert result.exit_code == EXIT_FAILURE
        assert "Workdir not found" in result.summary

    def test_returns_failure_when_tool_not_found(self, tmp_path: Path) -> None:
        """Test returns failure when tool executable not found."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        def raise_not_found(*args: Any, **kwargs: Any) -> None:
            raise FileNotFoundError("ruff not found")

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"ruff": raise_not_found}),
        ):
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        assert result.exit_code == EXIT_FAILURE
        assert "not found" in result.summary

    def test_mutmut_uses_config_timeout(self, tmp_path: Path) -> None:
        """Test mutmut runner receives timeout from config."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="mutmut",
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "mutmut"}
        mock_runner = MagicMock(return_value=mock_result)

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"mutmut": mock_runner}),
        ):
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"mutmut": {"timeout_minutes": 30}}},
            }
            cmd_run(args)

        # Verify timeout was passed (30 min * 60 = 1800 seconds)
        mock_runner.assert_called_once()
        _, call_args, _ = mock_runner.mock_calls[0]
        assert call_args[2] == 1800  # timeout in seconds

    def test_uses_workdir_from_config(self, tmp_path: Path) -> None:
        """Test uses workdir from repo.subdir in config."""
        subdir = tmp_path / "src"
        subdir.mkdir()

        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,  # Not specified
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "ruff"}
        mock_runner = MagicMock(return_value=mock_result)

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"ruff": mock_runner}),
        ):
            mock_load.return_value = {
                "language": "python",
                "repo": {"subdir": "src"},
            }
            cmd_run(args)

        # Verify the runner was called with the subdir path
        mock_runner.assert_called_once()
        call_args = mock_runner.call_args
        workdir_arg = call_args[0][0]
        assert str(workdir_arg).endswith("src")

    def test_writes_output_to_custom_path(self, tmp_path: Path) -> None:
        """Test writes output to custom path when specified."""
        output_file = tmp_path / "custom_output.json"
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=True,
            workdir=None,
            output=str(output_file),
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "ruff"}

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"ruff": MagicMock(return_value=mock_result)}),
        ):
            mock_load.return_value = {"language": "python"}
            cmd_run(args)

        # Verify write_json was called with custom path
        mock_result.write_json.assert_called_once()
        path_arg = mock_result.write_json.call_args[0][0]
        assert str(path_arg) == str(output_file)

    def test_success_returns_command_result_with_output_info(self, tmp_path: Path) -> None:
        """Test successful run returns CommandResult with output path in data."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=False,  # Doesn't matter now - always returns CommandResult
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=True,
        )

        mock_result = MagicMock(spec=ToolResult)
        mock_result.success = True
        mock_result.to_payload.return_value = {"tool": "ruff", "ran": True, "success": True}

        with (
            patch("cihub.commands.run.load_ci_config") as mock_load,
            patch.dict(RUNNERS, {"ruff": MagicMock(return_value=mock_result)}),
        ):
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        # Now returns CommandResult instead of int
        assert result.exit_code == EXIT_SUCCESS
        assert "items" in result.data
        assert any("Wrote output" in item for item in result.data["items"])

    def test_config_error_returns_command_result(self, tmp_path: Path) -> None:
        """Test config error returns CommandResult with problem details."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="ruff",
            json=False,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.side_effect = Exception("Config missing")
            result = cmd_run(args)

        # Returns CommandResult with error info
        assert result.exit_code == EXIT_FAILURE
        assert "Failed to load config" in result.summary
        assert any(p["code"] == "CIHUB-RUN-001" for p in result.problems)

    def test_unsupported_tool_returns_command_result(self, tmp_path: Path) -> None:
        """Test unsupported tool returns CommandResult with problem details."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="unknown_tool",
            json=False,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {"language": "python"}
            result = cmd_run(args)

        # Returns CommandResult with error info
        assert result.exit_code == EXIT_USAGE
        assert "Unsupported tool" in result.summary
        assert any(p["code"] == "CIHUB-RUN-003" for p in result.problems)

    def test_disabled_tool_returns_command_result(self, tmp_path: Path) -> None:
        """Test disabled tool returns CommandResult with skipped status."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="bandit",
            json=False,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"bandit": False}},
            }
            result = cmd_run(args)

        # Returns CommandResult indicating tool was skipped
        assert result.exit_code == EXIT_SUCCESS
        assert "skipped" in result.summary
        assert "items" in result.data

    def test_pip_audit_normalizes_to_pip_audit(self, tmp_path: Path) -> None:
        """Test pip-audit tool name is normalized to pip_audit internally."""
        args = argparse.Namespace(
            repo=str(tmp_path),
            tool="pip-audit",  # Hyphenated form
            json=True,
            workdir=None,
            output=None,
            output_dir=str(tmp_path / ".cihub"),
            force=False,
        )

        with patch("cihub.commands.run.load_ci_config") as mock_load:
            mock_load.return_value = {
                "language": "python",
                "python": {"tools": {"pip_audit": False}},  # Config uses underscore
            }
            result = cmd_run(args)

        # Should recognize pip-audit as pip_audit and skip it
        assert result.exit_code == EXIT_SUCCESS
        assert "skipped" in result.summary



"""Base class for language-specific CI strategies.

This module defines the LanguageStrategy abstract base class that encapsulates
all language-specific behavior for the CI engine. By implementing this interface,
new languages can be added without modifying existing code.

Design rationale (per CLEAN_CODE.md Part 2.1):
- Eliminates 46+ if-language conditionals scattered across 20 files
- Adding Go/Rust/TypeScript = one new file, zero modifications to existing code
- Testing: mock strategy, verify calls
- No more duplicate gate logic
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from cihub.core.ci_report import RunContext


class LanguageStrategy(ABC):
    """Abstract base class for language-specific CI behavior.

    Each language (Python, Java, etc.) implements this interface to provide:
    - Tool runner mappings
    - Default tool execution order
    - Gate evaluation logic
    - Language detection heuristics
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Language identifier (e.g., 'python', 'java').

        This is used as the key in the strategy registry and must match
        the language value in configuration files.
        """

    @abstractmethod
    def get_runners(self) -> dict[str, Callable[..., Any]]:
        """Return tool name -> runner function mapping.

        NOTE: This returns only tools with actual runner functions.
        Derived/virtual tools (e.g., hypothesis, jqwik) and external tools
        (e.g., codeql) are NOT included here - they are handled specially
        in run_tools() via delegation to existing _run_*_tools() functions.

        Returns:
            Dictionary mapping tool names (e.g., 'pytest', 'ruff') to their
            runner functions from cihub.core.ci_runner.
        """

    @abstractmethod
    def get_default_tools(self) -> list[str]:
        """Return all tools for this language (from registry).

        NOTE: This returns ALL tools in scope for the language, including:
        - Runnable tools (have entries in get_runners())
        - Derived/virtual tools (e.g., hypothesis mirrors pytest, jqwik mirrors build)
        - External tools (e.g., codeql, tracked via env vars)

        Returns:
            List of tool names from the tool registry for this language.
        """

    @abstractmethod
    def get_thresholds(self) -> tuple[Any, ...]:
        """Return threshold specifications for this language.

        Returns:
            Tuple of ThresholdSpec objects defining quality gates.
        """

    @abstractmethod
    def get_tool_specs(self) -> tuple[Any, ...]:
        """Return tool specifications for this language.

        Returns:
            Tuple of ToolSpec objects defining available tools.
        """

    @abstractmethod
    def run_tools(
        self,
        config: dict[str, Any],
        repo_path: Path,
        workdir: str,
        output_dir: Path,
        problems: list[dict[str, Any]],
    ) -> tuple[dict[str, dict[str, Any]], dict[str, bool], dict[str, bool]]:
        """Execute all enabled tools, return results.

        Args:
            config: CI configuration dictionary
            repo_path: Path to the repository root
            workdir: Working directory relative to repo_path
            output_dir: Directory for tool output artifacts
            problems: List to append warnings/errors to

        Returns:
            Tuple of (tool_outputs, tools_ran, tools_success) dictionaries.
        """

    @abstractmethod
    def evaluate_gates(
        self,
        report: dict[str, Any],
        thresholds: dict[str, Any],
        tools_configured: dict[str, bool],
        config: dict[str, Any],
    ) -> list[str]:
        """Evaluate quality gates, return list of failure messages.

        Args:
            report: The CI report with results and metrics
            thresholds: Effective threshold values
            tools_configured: Which tools are configured/enabled
            config: Full CI configuration

        Returns:
            List of gate failure messages. Empty list means all gates pass.
        """

    @abstractmethod
    def resolve_thresholds(self, config: dict[str, Any]) -> dict[str, Any]:
        """Resolve effective threshold values from configuration.

        Args:
            config: CI configuration dictionary

        Returns:
            Dictionary of threshold keys to their effective values.
        """

    @abstractmethod
    def build_report(
        self,
        config: dict[str, Any],
        tool_results: dict[str, dict[str, Any]],
        tools_configured: dict[str, bool],
        tools_ran: dict[str, bool],
        tools_success: dict[str, bool],
        thresholds: dict[str, Any],
        context: RunContext,
        *,
        tools_require_run: dict[str, bool] | None = None,
    ) -> dict[str, Any]:
        """Build the CI report for this language.

        Args:
            config: CI configuration dictionary
            tool_results: Results from tool execution
            tools_configured: Which tools are configured
            tools_ran: Which tools actually ran
            tools_success: Which tools succeeded
            thresholds: Effective threshold values
            context: Run context with metadata
            tools_require_run: Tools with require_run_or_fail policy

        Returns:
            The complete CI report dictionary.
        """

    def detect(self, repo_path: Path) -> float:
        """Return confidence 0.0-1.0 that this language applies to the repo.

        Override this method to implement language detection heuristics.
        The strategy with the highest confidence above 0.5 wins.

        Args:
            repo_path: Path to the repository root

        Returns:
            Confidence score between 0.0 and 1.0
        """
        return 0.0

    def get_allowed_kwargs(self) -> frozenset[str]:
        """Return the set of kwargs that run_tools() accepts.

        Override in subclasses to specify which kwargs are valid for run_tools().
        This enables safe filtering to prevent unexpected kwargs from being passed.

        Returns:
            Frozenset of allowed kwarg names
        """
        return frozenset()

    def get_run_kwargs(
        self,
        config: dict[str, Any],
        **caller_kwargs: Any,
    ) -> dict[str, Any]:
        """Get language-specific kwargs for run_tools(), merged with caller kwargs.

        Override this to provide language-specific parameters like install_deps (Python)
        or build_tool (Java). The base implementation filters to allowed kwargs.

        Args:
            config: CI configuration dictionary
            **caller_kwargs: Kwargs from the caller (e.g., install_deps from CLI)

        Returns:
            Merged dictionary of kwargs for run_tools(), filtered to allowed kwargs
        """
        allowed = self.get_allowed_kwargs()
        if not allowed:
            return dict(caller_kwargs)
        # Filter to only allowed kwargs to prevent unexpected parameters
        return {k: v for k, v in caller_kwargs.items() if k in allowed}

    def get_virtual_tools(self) -> list[str]:
        """Return tools that don't have runners but are tracked in reports.

        Virtual/derived tools mirror the results of their parent tools:
        - Python: 'hypothesis' mirrors 'pytest' (property-based testing)
        - Java: 'jqwik' mirrors 'build' (property-based testing)

        These tools are included in get_default_tools() but NOT in get_runners().
        This method makes that distinction explicit for documentation and safety.

        Returns:
            List of virtual tool names that have no runner
        """
        return []

    def get_docker_config(self, config: dict[str, Any]) -> dict[str, Any]:
        """Extract docker configuration for this language.

        Encapsulates language-specific docker config access. This moves the
        config[language].tools.docker extraction into the strategy.

        Args:
            config: CI configuration dictionary

        Returns:
            Dict with docker config keys: compose_file, health_endpoint, etc.
        """
        lang_cfg = config.get(self.name, {})
        if not isinstance(lang_cfg, dict):
            return {}
        tools_cfg = lang_cfg.get("tools", {})
        if not isinstance(tools_cfg, dict):
            return {}
        docker_cfg = tools_cfg.get("docker", {})
        if not isinstance(docker_cfg, dict):
            return {}
        return dict(docker_cfg)

    def get_docker_compose_default(self) -> str | None:
        """Get default docker-compose file for this language.

        Python defaults to 'docker-compose.yml', Java has no default.

        Returns:
            Default compose filename or None
        """
        return None

    def get_context_extras(
        self,
        config: dict[str, Any],
        workdir_path: Path,
    ) -> dict[str, Any]:
        """Get extra context fields for this language.

        Java adds build_tool and project_type. Python has none.

        Args:
            config: CI configuration dictionary
            workdir_path: Absolute path to working directory

        Returns:
            Dictionary of extra context fields
        """
        return {}



"""Config management module for CI/CD Hub."""

from __future__ import annotations

from cihub.config.io import (
    ensure_dirs,
    list_profiles,
    list_repos,
    load_defaults,
    load_profile,
    load_profile_strict,
    load_repo_config,
    load_yaml_file,
    save_repo_config,
    save_yaml_file,
)
from cihub.config.merge import (
    build_effective_config,
    deep_merge,
    get_effective_config_for_repo,
)
from cihub.config.normalize import normalize_config, normalize_tool_configs, tool_enabled
from cihub.config.paths import PathConfig
from cihub.config.schema import get_schema, validate_config

__all__ = [
    "PathConfig",
    "build_effective_config",
    "deep_merge",
    "ensure_dirs",
    "get_effective_config_for_repo",
    "get_schema",
    "list_profiles",
    "list_repos",
    "load_defaults",
    "load_profile",
    "load_profile_strict",
    "load_repo_config",
    "load_yaml_file",
    "normalize_config",
    "normalize_tool_configs",
    "save_repo_config",
    "save_yaml_file",
    "tool_enabled",
    "validate_config",
]



"""Tests for cihub.core.gate_specs module."""

from __future__ import annotations

import pytest

from cihub.core.gate_specs import (
    JAVA_THRESHOLDS,
    JAVA_TOOL_KEYS,
    JAVA_TOOLS,
    PYTHON_THRESHOLDS,
    PYTHON_TOOL_KEYS,
    PYTHON_TOOLS,
    Category,
    Comparator,
    ThresholdSpec,
    ToolSpec,
    evaluate_threshold,
    get_metric_value,
    get_threshold_spec_by_key,
    get_thresholds,
    get_tool_keys,
    get_tools,
    threshold_rows,
    tool_rows,
)


class TestThresholdSpec:
    """Tests for ThresholdSpec dataclass."""

    def test_threshold_spec_is_frozen(self) -> None:
        spec = ThresholdSpec(
            label="Test",
            key="test_key",
            unit="%",
            comparator=Comparator.GTE,
            metric_key="test_metric",
            failure_template="failed",
        )
        with pytest.raises(AttributeError):
            spec.label = "Modified"  # type: ignore[misc]


class TestToolSpec:
    """Tests for ToolSpec dataclass."""

    def test_tool_spec_is_frozen(self) -> None:
        spec = ToolSpec(Category.TESTING, "Test", "test")
        with pytest.raises(AttributeError):
            spec.label = "Modified"  # type: ignore[misc]


class TestGetThresholds:
    """Tests for get_thresholds function."""

    def test_returns_python_thresholds(self) -> None:
        result = get_thresholds("python")
        assert result == PYTHON_THRESHOLDS

    def test_returns_java_thresholds(self) -> None:
        result = get_thresholds("java")
        assert result == JAVA_THRESHOLDS

    def test_defaults_to_python_for_unknown(self) -> None:
        result = get_thresholds("unknown")
        assert result == PYTHON_THRESHOLDS


class TestGetTools:
    """Tests for get_tools function."""

    def test_returns_python_tools(self) -> None:
        result = get_tools("python")
        assert result == PYTHON_TOOLS

    def test_returns_java_tools(self) -> None:
        result = get_tools("java")
        assert result == JAVA_TOOLS


class TestGetToolKeys:
    """Tests for get_tool_keys function."""

    def test_returns_python_tool_keys(self) -> None:
        result = get_tool_keys("python")
        assert result == PYTHON_TOOL_KEYS

    def test_returns_java_tool_keys(self) -> None:
        result = get_tool_keys("java")
        assert result == JAVA_TOOL_KEYS

    def test_python_tool_keys_include_sbom_and_hypothesis(self) -> None:
        keys = get_tool_keys("python")
        assert "sbom" in keys
        assert "hypothesis" in keys

    def test_java_tool_keys_include_sbom_but_exclude_build_placeholder(self) -> None:
        keys = get_tool_keys("java")
        assert "sbom" in keys
        assert "__build__" not in keys


class TestThresholdRows:
    """Tests for threshold_rows function."""

    def test_returns_tuple_format(self) -> None:
        rows = threshold_rows("python")
        assert isinstance(rows, list)
        assert len(rows) > 0
        # Each row should be (label, key, unit)
        for row in rows:
            assert len(row) == 3
            assert isinstance(row[0], str)
            assert isinstance(row[1], str)
            assert isinstance(row[2], str)

    def test_contains_expected_python_thresholds(self) -> None:
        rows = threshold_rows("python")
        keys = [row[1] for row in rows]
        assert "coverage_min" in keys
        assert "mutation_score_min" in keys
        assert "trivy_cvss_fail" in keys
        assert "max_ruff_errors" in keys

    def test_contains_expected_java_thresholds(self) -> None:
        rows = threshold_rows("java")
        keys = [row[1] for row in rows]
        assert "coverage_min" in keys
        assert "owasp_cvss_fail" in keys
        assert "max_checkstyle_errors" in keys
        assert "max_spotbugs_bugs" in keys


class TestToolRows:
    """Tests for tool_rows function."""

    def test_returns_tuple_format(self) -> None:
        rows = tool_rows("python")
        assert isinstance(rows, list)
        assert len(rows) > 0
        # Each row should be (category, label, key)
        for row in rows:
            assert len(row) == 3
            assert isinstance(row[0], str)
            assert isinstance(row[1], str)
            assert isinstance(row[2], str)

    def test_contains_expected_python_tools(self) -> None:
        rows = tool_rows("python")
        keys = [row[2] for row in rows]
        assert "pytest" in keys
        assert "ruff" in keys
        assert "bandit" in keys
        assert "trivy" in keys

    def test_contains_expected_java_tools(self) -> None:
        rows = tool_rows("java")
        keys = [row[2] for row in rows]
        assert "__build__" in keys
        assert "jacoco" in keys
        assert "checkstyle" in keys
        assert "owasp" in keys


class TestEvaluateThreshold:
    """Tests for evaluate_threshold function."""

    def test_gte_comparator_passes_when_above(self) -> None:
        spec = ThresholdSpec(
            label="Coverage",
            key="coverage_min",
            unit="%",
            comparator=Comparator.GTE,
            metric_key="coverage",
            failure_template="coverage {value}% < {threshold}%",
        )
        passed, msg = evaluate_threshold(spec, 80, 70)
        assert passed is True
        assert msg is None

    def test_gte_comparator_fails_when_below(self) -> None:
        spec = ThresholdSpec(
            label="Coverage",
            key="coverage_min",
            unit="%",
            comparator=Comparator.GTE,
            metric_key="coverage",
            failure_template="coverage {value}% < {threshold}%",
        )
        passed, msg = evaluate_threshold(spec, 60, 70)
        assert passed is False
        assert msg == "coverage 60% < 70%"

    def test_lte_comparator_passes_when_below(self) -> None:
        spec = ThresholdSpec(
            label="Max Errors",
            key="max_errors",
            unit="",
            comparator=Comparator.LTE,
            metric_key="errors",
            failure_template="errors {value} > {threshold}",
        )
        passed, msg = evaluate_threshold(spec, 5, 10)
        assert passed is True
        assert msg is None

    def test_lte_comparator_fails_when_above(self) -> None:
        spec = ThresholdSpec(
            label="Max Errors",
            key="max_errors",
            unit="",
            comparator=Comparator.LTE,
            metric_key="errors",
            failure_template="errors {value} > {threshold}",
        )
        passed, msg = evaluate_threshold(spec, 15, 10)
        assert passed is False
        assert msg == "errors 15 > 10"

    def test_cvss_comparator_passes_when_below(self) -> None:
        spec = ThresholdSpec(
            label="CVSS Fail",
            key="cvss_fail",
            unit="",
            comparator=Comparator.CVSS,
            metric_key="max_cvss",
            failure_template="max CVSS {value:.1f} >= {threshold:.1f}",
        )
        passed, msg = evaluate_threshold(spec, 6.5, 7.0)
        assert passed is True
        assert msg is None

    def test_cvss_comparator_fails_when_at_or_above(self) -> None:
        spec = ThresholdSpec(
            label="CVSS Fail",
            key="cvss_fail",
            unit="",
            comparator=Comparator.CVSS,
            metric_key="max_cvss",
            failure_template="max CVSS {value:.1f} >= {threshold:.1f}",
        )
        passed, msg = evaluate_threshold(spec, 7.5, 7.0)
        assert passed is False
        assert msg == "max CVSS 7.5 >= 7.0"

    def test_cvss_comparator_skipped_when_threshold_zero(self) -> None:
        spec = ThresholdSpec(
            label="CVSS Fail",
            key="cvss_fail",
            unit="",
            comparator=Comparator.CVSS,
            metric_key="max_cvss",
            failure_template="max CVSS {value:.1f} >= {threshold:.1f}",
        )
        passed, msg = evaluate_threshold(spec, 9.0, 0.0)
        assert passed is True
        assert msg is None


class TestGetThresholdSpecByKey:
    """Tests for get_threshold_spec_by_key function."""

    def test_finds_python_coverage_min(self) -> None:
        spec = get_threshold_spec_by_key("python", "coverage_min")
        assert spec is not None
        assert spec.key == "coverage_min"
        assert spec.comparator == Comparator.GTE

    def test_finds_java_owasp_cvss_fail(self) -> None:
        spec = get_threshold_spec_by_key("java", "owasp_cvss_fail")
        assert spec is not None
        assert spec.key == "owasp_cvss_fail"
        assert spec.comparator == Comparator.CVSS

    def test_returns_none_for_unknown_key(self) -> None:
        spec = get_threshold_spec_by_key("python", "nonexistent_key")
        assert spec is None


class TestGetMetricValue:
    """Tests for get_metric_value function."""

    def test_gets_value_from_tool_metrics(self) -> None:
        spec = ThresholdSpec(
            label="Errors",
            key="max_errors",
            unit="",
            comparator=Comparator.LTE,
            metric_key="ruff_errors",
            failure_template="errors {value}",
        )
        value = get_metric_value(spec, {}, {"ruff_errors": 5})
        assert value == 5

    def test_gets_value_from_results(self) -> None:
        spec = ThresholdSpec(
            label="Coverage",
            key="coverage_min",
            unit="%",
            comparator=Comparator.GTE,
            metric_key="coverage",
            failure_template="coverage {value}%",
        )
        value = get_metric_value(spec, {"coverage": 85}, {})
        assert value == 85

    def test_returns_zero_for_missing(self) -> None:
        spec = ThresholdSpec(
            label="Test",
            key="test",
            unit="",
            comparator=Comparator.LTE,
            metric_key="missing",
            failure_template="test",
        )
        value = get_metric_value(spec, {}, {})
        assert value == 0

    def test_converts_string_to_int(self) -> None:
        spec = ThresholdSpec(
            label="Test",
            key="test",
            unit="",
            comparator=Comparator.LTE,
            metric_key="count",
            failure_template="test",
        )
        value = get_metric_value(spec, {}, {"count": "10"})
        assert value == 10

    def test_converts_string_to_float(self) -> None:
        spec = ThresholdSpec(
            label="Test",
            key="test",
            unit="",
            comparator=Comparator.CVSS,
            metric_key="score",
            failure_template="test",
        )
        value = get_metric_value(spec, {}, {"score": "7.5"})
        assert value == 7.5


class TestRegistryConsistency:
    """Tests to ensure registry definitions are consistent."""

    def test_python_thresholds_have_required_fields(self) -> None:
        for spec in PYTHON_THRESHOLDS:
            assert spec.label
            assert spec.key
            assert spec.metric_key
            assert spec.failure_template

    def test_java_thresholds_have_required_fields(self) -> None:
        for spec in JAVA_THRESHOLDS:
            assert spec.label
            assert spec.key
            assert spec.metric_key
            assert spec.failure_template

    def test_python_tools_have_valid_categories(self) -> None:
        for spec in PYTHON_TOOLS:
            assert isinstance(spec.category, Category)
            assert spec.label
            assert spec.key

    def test_java_tools_have_valid_categories(self) -> None:
        for spec in JAVA_TOOLS:
            assert isinstance(spec.category, Category)
            assert spec.label
            assert spec.key

    def test_tool_keys_are_subset_of_tool_specs(self) -> None:
        python_spec_keys = {s.key for s in PYTHON_TOOLS}
        for key in PYTHON_TOOL_KEYS:
            # Tool keys should exist in tools OR be internal (like build)
            assert key in python_spec_keys or key in ["build"]

        java_spec_keys = {s.key for s in JAVA_TOOLS}
        for key in JAVA_TOOL_KEYS:
            assert key in java_spec_keys or key in ["build", "__build__"]



"""Backward compatibility shim for registry command.

This module re-exports all symbols from the new registry package
to maintain backward compatibility with existing imports like:
    from cihub.commands.registry_cmd import cmd_registry

The implementation has been moved to cihub.commands.registry/
"""

from __future__ import annotations

# Re-export everything from the new registry package
from cihub.commands.registry import (
    _cmd_add,
    _cmd_bootstrap,
    _cmd_diff,
    _cmd_export,
    _cmd_import,
    _cmd_list,
    _cmd_remove,
    _cmd_set,
    _cmd_show,
    _cmd_sync,
    _derive_hub_root_from_configs_dir,
    cmd_registry,
)

__all__ = [
    "cmd_registry",
    "_derive_hub_root_from_configs_dir",
    "_cmd_list",
    "_cmd_show",
    "_cmd_set",
    "_cmd_add",
    "_cmd_remove",
    "_cmd_diff",
    "_cmd_sync",
    "_cmd_bootstrap",
    "_cmd_export",
    "_cmd_import",
]
